from io import BytesIO

import hmac
import hashlib

from buidl.helper import (
    big_endian_to_int,
    encode_base58_checksum,
    hash160,
    hash256,
    int_to_big_endian,
    raw_decode_base58,
)


class FieldElement:
    def __init__(self, num, prime):
        if num >= prime or num < 0:
            error = "Num {} not in field range 0 to {}".format(num, prime - 1)
            raise ValueError(error)
        self.num = num
        self.prime = prime

    def __eq__(self, other):
        if other is None:
            return False
        return self.num == other.num and self.prime == other.prime

    def __ne__(self, other):
        # this should be the inverse of the == operator
        return not (self == other)

    def __repr__(self):
        return "FieldElement_{}({})".format(self.prime, self.num)

    def __add__(self, other):
        if self.prime != other.prime:
            raise TypeError("Cannot add two numbers in different Fields")
        # self.num and other.num are the actual values
        num = (self.num + other.num) % self.prime
        # self.prime is what you'll need to mod against
        prime = self.prime
        # You need to return an element of the same class
        # use: self.__class__(num, prime)
        return self.__class__(num, prime)

    def __sub__(self, other):
        if self.prime != other.prime:
            raise TypeError("Cannot add two numbers in different Fields")
        # self.num and other.num are the actual values
        num = (self.num - other.num) % self.prime
        # self.prime is what you'll need to mod against
        prime = self.prime
        # You need to return an element of the same class
        # use: self.__class__(num, prime)
        return self.__class__(num, prime)

    def __mul__(self, other):
        if self.prime != other.prime:
            raise TypeError("Cannot add two numbers in different Fields")
        # self.num and other.num are the actual values
        num = (self.num * other.num) % self.prime
        # self.prime is what you'll need to mod against
        prime = self.prime
        # You need to return an element of the same class
        # use: self.__class__(num, prime)
        return self.__class__(num, prime)

    def __pow__(self, n):
        # remember Fermat's Little Theorem:
        # self.num**(p-1) % p == 1
        # you might want to use % operator on n
        prime = self.prime
        num = pow(self.num, n % (prime - 1), prime)
        return self.__class__(num, prime)

    def __truediv__(self, other):
        if self.prime != other.prime:
            raise TypeError("Cannot add two numbers in different Fields")
        # self.num and other.num are the actual values
        num = (self.num * pow(other.num, self.prime - 2, self.prime)) % self.prime
        # self.prime is what you'll need to mod against
        prime = self.prime
        # use fermat's little theorem:
        # self.num**(p-1) % p == 1
        # this means:
        # 1/n == pow(n, p-2, p)
        # You need to return an element of the same class
        # use: self.__class__(num, prime)
        return self.__class__(num, prime)

    def __rmul__(self, coefficient):
        num = (self.num * coefficient) % self.prime
        return self.__class__(num=num, prime=self.prime)


class Point:
    def __init__(self, x, y, a, b):
        self.a = a
        self.b = b
        self.x = x
        self.y = y
        # x being None and y being None represents the point at infinity
        # Check for that here since the equation below won't make sense
        # with None values for both.
        if self.x is None and self.y is None:
            return
        # make sure that the elliptic curve equation is satisfied
        # y**2 == x**3 + a*x + b
        if self.y ** 2 != self.x ** 3 + a * x + b:
            # if not, raise a ValueError
            raise ValueError("({}, {}) is not on the curve".format(self.x, self.y))

    def __eq__(self, other):
        return (
            self.x == other.x
            and self.y == other.y
            and self.a == other.a
            and self.b == other.b
        )

    def __ne__(self, other):
        # this should be the inverse of the == operator
        return not (self == other)

    def __repr__(self):
        if self.x is None:
            return "Point(infinity)"
        else:
            return "Point({},{})_{}".format(self.x.num, self.y.num, self.x.prime)

    def __add__(self, other):
        if self.a != other.a or self.b != other.b:
            raise TypeError(
                "Points {}, {} are not on the same curve".format(self, other)
            )
        # Case 0.0: self is the point at infinity, return other
        if self.x is None:
            return other
        # Case 0.1: other is the point at infinity, return self
        if other.x is None:
            return self

        # Case 1: self.x == other.x, self.y != other.y
        # Result is point at infinity
        if self.x == other.x and self.y != other.y:
            # Remember to return an instance of this class:
            # self.__class__(x, y, a, b)
            return self.__class__(None, None, self.a, self.b)

        # Case 2: self.x != other.x
        if self.x != other.x:
            # Formula (x3,y3)==(x1,y1)+(x2,y2)
            # s=(y2-y1)/(x2-x1)
            s = (other.y - self.y) / (other.x - self.x)
            # x3=s**2-x1-x2
            x = s ** 2 - self.x - other.x
            # y3=s*(x1-x3)-y1
            y = s * (self.x - x) - self.y
            return self.__class__(x, y, self.a, self.b)

        # Case 3: self.x == other.x, self.y == other.y
        else:
            # Formula (x3,y3)=(x1,y1)+(x1,y1)
            # s=(3*x1**2+a)/(2*y1)
            s = (3 * self.x ** 2 + self.a) / (2 * self.y)
            # x3=s**2-2*x1
            x = s ** 2 - 2 * self.x
            # y3=s*(x1-x3)-y1
            y = s * (self.x - x) - self.y
            return self.__class__(x, y, self.a, self.b)

    def __rmul__(self, coefficient):
        # rmul calculates coefficient * self
        coef = coefficient
        current = self
        # start at 0
        result = self.__class__(None, None, self.a, self.b)
        while coef:
            # if the bit at this binary expansion is 1, add
            if coef & 1:
                result += current
            # double the point
            current += current
            coef >>= 1
        return result


A = 0
B = 7
P = 2 ** 256 - 2 ** 32 - 977
N = 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFEBAAEDCE6AF48A03BBFD25E8CD0364141


class S256Field(FieldElement):
    def __init__(self, num, prime=None):
        super().__init__(num=num, prime=P)

    def hex(self):
        return "{:x}".format(self.num).zfill(64)

    def __repr__(self):
        return self.hex()

    def sqrt(self):
        return self ** ((P + 1) // 4)


class S256Point(Point):
    def __init__(self, x, y, a=None, b=None):
        a, b = S256Field(A), S256Field(B)
        if type(x) == int:
            super().__init__(x=S256Field(x), y=S256Field(y), a=a, b=b)
        else:
            super().__init__(x=x, y=y, a=a, b=b)

    def __eq__(self, other):
        return self.x == other.x and self.y == other.y

    def __repr__(self):
        if self.x is None:
            return "S256Point(infinity)"
        else:
            return "S256Point({},{})".format(hex(self.x.num), hex(self.y.num))

    def __rmul__(self, coefficient):
        # we want to mod by N to make this simple
        coef = coefficient % N
        return super().__rmul__(coef)

    def __add__(self, other):
        """If other is an int, multiplies scalar by generator, adds result to current point"""
        if type(other) == int:
            return super().__add__(other * G)
        else:
            return super().__add__(other)

    def sec(self, compressed=True):
        # returns the binary version of the sec format, NOT hex
        # if compressed, starts with b'\x02' if self.y.num is even, b'\x03' if self.y is odd
        # then self.x.num
        # remember, you have to convert self.x.num/self.y.num to binary using int_to_big_endian
        x = int_to_big_endian(self.x.num, 32)
        if compressed:
            if self.y.num % 2 == 0:
                return b"\x02" + x
            else:
                return b"\x03" + x
        else:
            # if non-compressed, starts with b'\x04' followod by self.x and then self.y
            y = int_to_big_endian(self.y.num, 32)
            return b"\x04" + x + y

    def hash160(self, compressed=True):
        # get the sec
        sec = self.sec(compressed)
        # hash160 the sec
        return hash160(sec)

    def p2pkh_script(self, compressed=True):
        """Returns the p2pkh Script object"""
        h160 = self.hash160(compressed)
        # avoid circular dependency
        from buidl.script import P2PKHScriptPubKey

        return P2PKHScriptPubKey(h160)

    def p2wpkh_script(self):
        """Returns the p2wpkh Script object"""
        h160 = self.hash160(True)
        # avoid circular dependency
        from buidl.script import P2WPKHScriptPubKey

        return P2WPKHScriptPubKey(h160)

    def p2sh_p2wpkh_redeem_script(self):
        """Returns the RedeemScript for a p2sh-p2wpkh redemption"""
        return self.p2wpkh_script().redeem_script()

    def address(self, compressed=True, network="mainnet"):
        """Returns the p2pkh address string"""
        return self.p2pkh_script(compressed).address(network)

    def bech32_address(self, network="mainnet"):
        """Returns the p2wpkh bech32 address string"""
        return self.p2wpkh_script().address(network)

    def p2sh_p2wpkh_address(self, network="mainnet"):
        """Returns the p2sh-p2wpkh base58 address string"""
        return self.p2wpkh_script().p2sh_address(network)

    def verify(self, z, sig):
        # remember sig.r and sig.s are the main things we're checking
        # remember 1/s = pow(s, N-2, N)
        s_inv = pow(sig.s, N - 2, N)
        # u = z / s
        u = z * s_inv % N
        # v = r / s
        v = sig.r * s_inv % N
        # u*G + v*P should have as the x coordinate, r
        total = u * G + v * self
        return total.x.num == sig.r

    def verify_message(self, message, sig):
        """Verify a message in the form of bytes. Assumes that the z
        is calculated using hash256 interpreted as a big-endian integer"""
        # calculate the hash256 of the message
        h256 = hash256(message)
        # z is the big-endian interpretation. use big_endian_to_int
        z = big_endian_to_int(h256)
        # verify the message using the self.verify method
        return self.verify(z, sig)

    @classmethod
    def parse(self, sec_bin):
        """returns a Point object from a compressed sec binary (not hex)"""
        if sec_bin[0] == 4:
            x = int(sec_bin[1:33].hex(), 16)
            y = int(sec_bin[33:65].hex(), 16)
            return S256Point(x=x, y=y)
        is_even = sec_bin[0] == 2
        x = S256Field(int(sec_bin[1:].hex(), 16))
        # right side of the equation y^2 = x^3 + 7
        alpha = x ** 3 + S256Field(B)
        # solve for left side
        beta = alpha.sqrt()
        if beta.num % 2 == 0:
            even_beta = beta
            odd_beta = S256Field(P - beta.num)
        else:
            even_beta = S256Field(P - beta.num)
            odd_beta = beta
        if is_even:
            return S256Point(x, even_beta)
        else:
            return S256Point(x, odd_beta)


G = S256Point(
    0x79BE667EF9DCBBAC55A06295CE870B07029BFCDB2DCE28D959F2815B16F81798,
    0x483ADA7726A3C4655DA4FBFC0E1108A8FD17B448A68554199C47D08FFB10D4B8,
)


class Signature:
    def __init__(self, r, s):
        self.r = r
        self.s = s

    def __repr__(self):
        return "Signature({:x},{:x})".format(self.r, self.s)

    def der(self):
        # convert the r part to bytes
        rbin = int_to_big_endian(self.r, 32)
        # if rbin has a high bit, add a 00
        if rbin[0] >= 128:
            rbin = b"\x00" + rbin
        while rbin[0] == 0:
            if rbin[1] >= 128:
                break
            else:
                rbin = rbin[1:]
        result = bytes([2, len(rbin)]) + rbin
        sbin = int_to_big_endian(self.s, 32)
        # if sbin has a high bit, add a 00
        if sbin[0] >= 128:
            sbin = b"\x00" + sbin
        while sbin[0] == 0:
            if sbin[1] >= 128:
                break
            else:
                sbin = sbin[1:]
        result += bytes([2, len(sbin)]) + sbin
        return bytes([0x30, len(result)]) + result

    @classmethod
    def parse(cls, signature_bin):
        s = BytesIO(signature_bin)
        compound = s.read(1)[0]
        if compound != 0x30:
            raise RuntimeError("Bad Signature")
        length = s.read(1)[0]
        if length + 2 != len(signature_bin):
            raise RuntimeError("Bad Signature Length")
        marker = s.read(1)[0]
        if marker != 0x02:
            raise RuntimeError("Bad Signature")
        rlength = s.read(1)[0]
        r = int(s.read(rlength).hex(), 16)
        marker = s.read(1)[0]
        if marker != 0x02:
            raise RuntimeError("Bad Signature")
        slength = s.read(1)[0]
        s = int(s.read(slength).hex(), 16)
        if len(signature_bin) != 6 + rlength + slength:
            raise RuntimeError("Signature too long")
        return cls(r, s)


class PrivateKey:
    def __init__(self, secret, network="mainnet", compressed=True):
        self.secret = secret
        self.point = secret * G
        self.network = network
        self.compressed = compressed

    def hex(self):
        return "{:x}".format(self.secret).zfill(64)

    def sign(self, z):
        # we need use deterministic k
        k = self.deterministic_k(z)
        # r is the x coordinate of the resulting point k*G
        r = (k * G).x.num
        # remember 1/k = pow(k, N-2, N)
        k_inv = pow(k, N - 2, N)
        # s = (z+r*secret) / k
        s = (z + r * self.secret) * k_inv % N
        if s > N / 2:
            s = N - s
        # return an instance of Signature:
        # Signature(r, s)
        return Signature(r, s)

    def deterministic_k(self, z):
        k = b"\x00" * 32
        v = b"\x01" * 32
        if z > N:
            z -= N
        z_bytes = int_to_big_endian(z, 32)
        secret_bytes = int_to_big_endian(self.secret, 32)
        s256 = hashlib.sha256
        k = hmac.new(k, v + b"\x00" + secret_bytes + z_bytes, s256).digest()
        v = hmac.new(k, v, s256).digest()
        k = hmac.new(k, v + b"\x01" + secret_bytes + z_bytes, s256).digest()
        v = hmac.new(k, v, s256).digest()
        while True:
            v = hmac.new(k, v, s256).digest()
            candidate = big_endian_to_int(v)
            if candidate >= 1 and candidate < N:
                return candidate
            k = hmac.new(k, v + b"\x00", s256).digest()
            v = hmac.new(k, v, s256).digest()

    def sign_message(self, message):
        """Sign a message in the form of bytes instead of the z. The z should
        be assumed to be the hash256 of the message interpreted as a big-endian
        integer."""
        # compute the hash256 of the message
        h256 = hash256(message)
        # z is the big-endian interpretation. use big_endian_to_int
        z = big_endian_to_int(h256)
        # sign the message using the self.sign method
        return self.sign(z)

    def wif(self, compressed=True):
        # convert the secret from integer to a 32-bytes in big endian using int_to_big_endian(x, 32)
        secret_bytes = int_to_big_endian(self.secret, 32)
        # prepend b'\xef' on testnet/signet, b'\x80' on mainnet
        if self.network == "mainnet":
            prefix = b"\x80"
        else:
            prefix = b"\xef"
        # append b'\x01' if compressed
        if compressed:
            suffix = b"\x01"
        else:
            suffix = b""
        # encode_base58_checksum the whole thing
        return encode_base58_checksum(prefix + secret_bytes + suffix)

    @classmethod
    def parse(cls, wif):
        """Converts WIF to a PrivateKey object"""
        raw = raw_decode_base58(wif)
        if len(raw) == 34:
            compressed = True
            if raw[-1] != 1:
                raise ValueError("Invalid WIF")
            raw = raw[:-1]
        else:
            compressed = False
        secret = big_endian_to_int(raw[1:])
        if raw[0] == 0xEF:
            network = "testnet"
        elif raw[0] == 0x80:
            network = "mainnet"
        else:
            raise ValueError("Invalid WIF")
        return cls(secret, network=network, compressed=compressed)



from unittest import TestCase

from buidl.blinding import blind_xpub, combine_bip32_paths, secure_secret_path
from buidl.hd import HDPrivateKey, is_valid_bip32_path


class BlindingTest(TestCase):
    def test_generate_secure_secret_path(self):
        for depth in range(1, 8):
            res = secure_secret_path(depth=depth)
            self.assertTrue(is_valid_bip32_path(res))
            self.assertEqual(depth, res.count("/"))

    def test_generate_secure_secret_path_fail(self):
        with self.assertRaises(ValueError):
            secure_secret_path(depth="foo")
            secure_secret_path(depth=-1)
            secure_secret_path(depth=0)
            secure_secret_path(depth=257)

    def test_combine_paths(self):
        self.assertEqual(combine_bip32_paths("m/1", "m/2/3"), "m/1/2/3")
        self.assertEqual(combine_bip32_paths("m/1h", "m/2/3"), "m/1h/2/3")
        self.assertEqual(
            combine_bip32_paths("m/48h/1h/0h/2h", "m/1/2/3/4/5"),
            "m/48h/1h/0h/2h/1/2/3/4/5",
        )
        self.assertEqual(combine_bip32_paths("m", "m/1"), "m/1")
        self.assertEqual(combine_bip32_paths("m/1", "m"), "m/1")
        self.assertEqual(combine_bip32_paths("m", "m"), "m")

    def test_combine_paths_error(self):
        with self.assertRaises(ValueError):
            combine_bip32_paths("m/foo", "m/1/2/3")
            combine_bip32_paths("m/1/2/3", "m/foo")

    def test_blind_m48_vpub(self):
        # All test vectors below compared manually to https://iancoleman.io/bip39/ and then converted with https://jlopp.github.io/xpub-converter/
        starting_path = "m/48h/1h/0h/2h"
        # Vpub version bytes
        starting_vpub = (
            HDPrivateKey.from_mnemonic("oil " * 12)
            .traverse(starting_path)
            .xpub(bytes.fromhex("02575483"))
        )
        self.assertEqual(
            starting_vpub,
            "Vpub5mvQbnmqKfpPjWfAZEw5Xjdr6UjnjyZEirzrhNMSuKjL8Qfd3nqLBkrBrVXNeMgKCjPXbyLnSCn6qcD8fHQCkNnNLnkpQtY3sh4MHmywvbe",
        )
        secret_path = "m/920870093/318569592/821713943/1914815254/1398142787/9"  # randomly generated
        have = blind_xpub(
            starting_xpub=starting_vpub,
            starting_path=starting_path,
            secret_path=secret_path,
        )
        want = {
            "blinded_full_path": "m/48h/1h/0h/2h/920870093/318569592/821713943/1914815254/1398142787/9",
            "blinded_child_xpub": "Vpub5xWzaq6Ya7K9Y2UeFyL8SfHhydnp5FZDskm9mkAfUppkXVJDKrnLxU2Ezd55k8RSzSf4YETJj982NJQGCSzKJxUa6oQmbe1HWTRavCRzzxj",
        }
        self.assertEqual(have, want)

    def test_blind_m48_xpub(self):
        # All test vectors below compared manually to https://iancoleman.io/bip39/ and then converted with https://jlopp.github.io/xpub-converter/
        starting_path = "m/48h/0h/0h/2h"
        starting_xpub = (
            HDPrivateKey.from_mnemonic("oil " * 12).traverse(starting_path).xpub()
        )

        self.assertEqual(
            starting_xpub,
            "xpub6F8WgTkiV8iDPFG1Kv4sNrcBNMMgKK4cjfxjdZWvR3kChfbt3L2dJF7xmCHBMGMmxjyzwgjdFkh9UN3623YpsmqN1KwZGR45Y3ANLQQX87u",
        )
        secret_path = "m/920870093/318569592/821713943/1914815254/1398142787/9"  # randomly generated
        have = blind_xpub(
            starting_xpub=starting_xpub,
            starting_path=starting_path,
            secret_path=secret_path,
        )
        want = {
            "blinded_full_path": "m/48h/0h/0h/2h/920870093/318569592/821713943/1914815254/1398142787/9",
            "blinded_child_xpub": "xpub6RKvkus7fgUnP2trNVo2N1jaQeGPBHCV9m63Jaje8EW2Ry5KjYySb1tbPSi3E7Vh6ZzxRn15hUmBg5KmSaQvKjZmTvXKQnRPXcoJS9PXkiS",
        }
        self.assertEqual(have, want)

    def test_blind_root_xpub(self):
        # All test vectors below compared manually to https://iancoleman.io/bip39/

        starting_path = "m"
        # The .traverse() here does nothing and is optional, just here for consistency/clarity:
        root_xpub = (
            HDPrivateKey.from_mnemonic("bacon " * 24).traverse(starting_path).xpub()
        )

        self.assertEqual(
            root_xpub,
            "xpub661MyMwAqRbcGMzZtBZrKWaDMAQKjd5bMSnHr8qBfCz5zMwZEuajnc8cxjsEUECAZDeDC7s4zbT3Z9KrrM9wJ3MaT6sH9eRYxLY1BNf45BF",
        )
        secret_path = "m/1/2/3/4/5/6/7/8/9"
        have = blind_xpub(
            starting_xpub=root_xpub, starting_path="m", secret_path=secret_path
        )
        want = {
            "blinded_full_path": secret_path,  # the same since we started with the root path
            "blinded_child_xpub": "xpub6QbmAk7amacgfQVY3aTAiqhM1okb4Gi3EFzBJQMEja1uU4gZvG9vwiExr2C1ryZqnyczTvvaNhw63UALmXGkFo7FHhajX8SNimgK3zu3tp5",
        }
        self.assertEqual(have, want)

        starting_path = "m/48h/0h/0h/2h"
        child_xpub = (
            HDPrivateKey.from_mnemonic("bacon " * 24).traverse(starting_path).xpub()
        )
        self.assertEqual(
            child_xpub,
            "xpub6EeqK2JLwngrHJEQ4X4iqrySZV9qU3TgwMgf6NStLZa37AfNiHTtTE9ji1F9YQDLArJMLy8sw3Q2samVj5VQQjaaUHr5z2Hz57NWHJCfh31",
        )
        secret_path = "m/920870093/318569592/821713943/1914815254/1398142787/9"  # randomly generated
        have = blind_xpub(
            starting_xpub=child_xpub,
            starting_path=starting_path,
            secret_path=secret_path,
        )
        want = {
            "blinded_full_path": "m/48h/0h/0h/2h/920870093/318569592/821713943/1914815254/1398142787/9",
            "blinded_child_xpub": "xpub6SDyub38LYeUd211VxRZCiVnU4fSr4ZwPssFGMxyyXjZ6EvrA4ZkZhp4f9My2qzqSJFHxkAv8ctaGipx7UM6sCzmso7wxiohHBhcBs7ipWz",
        }
        self.assertEqual(have, want)

    def test_bad_depth(self):
        starting_path = "m/48h/0h/0h/2h"
        starting_xpub = (
            HDPrivateKey.from_mnemonic("bacon " * 24).traverse(starting_path).xpub()
        )
        with self.assertRaises(ValueError):
            blind_xpub(
                starting_xpub=starting_xpub, starting_path="m/1", secret_path="m/999"
            )



from unittest import TestCase

from buidl.bech32 import (
    encode_bech32_checksum,
    decode_bech32,
    BECH32_ALPHABET,
)


class Bech32Test(TestCase):
    def test_bech32(self):
        tests = [
            {
                "hex_script": "00201863143c14c5166804bd19203356da136c985678cd4d27a1b8c6329604903262",
                "mainnet": "bc1qrp33g0q5c5txsp9arysrx4k6zdkfs4nce4xj0gdcccefvpysxf3qccfmv3",
                "testnet": "tb1qrp33g0q5c5txsp9arysrx4k6zdkfs4nce4xj0gdcccefvpysxf3q0sl5k7",
                "signet": "tb1qrp33g0q5c5txsp9arysrx4k6zdkfs4nce4xj0gdcccefvpysxf3q0sl5k7",
            },
            {
                "hex_script": "00200000000000000000000000000000000000000000000000000000000000000000",
                "mainnet": "bc1qqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqthqst8",
                "testnet": "tb1qqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqulkl3g",
                "signet": "tb1qqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqulkl3g",
            },
            {
                "hex_script": "00140000000000000000000000000000000000000000",
                "mainnet": "bc1qqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqq9e75rs",
                "testnet": "tb1qqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqq0l98cr",
                "signet": "tb1qqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqq0l98cr",
            },
            {
                "hex_script": "0020000000c4a5cad46221b2a187905e5266362b99d5e91c6ce24d165dab93e86433",
                "mainnet": "bc1qqqqqp399et2xygdj5xreqhjjvcmzhxw4aywxecjdzew6hylgvses5wp4dt",
                "testnet": "tb1qqqqqp399et2xygdj5xreqhjjvcmzhxw4aywxecjdzew6hylgvsesrxh6hy",
                "signet": "tb1qqqqqp399et2xygdj5xreqhjjvcmzhxw4aywxecjdzew6hylgvsesrxh6hy",
            },
        ]
        for test in tests:
            raw = bytes.fromhex(test["hex_script"])
            want = test["signet"]
            version = BECH32_ALPHABET.index(want[3:4])
            result = encode_bech32_checksum(raw, network="signet")
            self.assertEqual(result, want)
            got_network, got_version, got_raw = decode_bech32(result)
            self.assertEqual(got_network, "testnet")
            self.assertEqual(got_version, version)
            self.assertEqual(got_raw, raw[2:])
            for network in ("mainnet", "testnet"):
                want = test[network]
                version = BECH32_ALPHABET.index(want[3:4])
                result = encode_bech32_checksum(raw, network=network)
                self.assertEqual(result, want)
                got_network, got_version, got_raw = decode_bech32(result)
                self.assertEqual(got_network, network)
                self.assertEqual(got_version, version)
                self.assertEqual(got_raw, raw[2:])



from buidl.bech32 import (
    cbor_encode,
    cbor_decode,
    bc32encode,
    bc32decode,
    uses_only_bech32_chars,
)
from buidl.helper import is_intable

from binascii import a2b_base64, b2a_base64
from math import ceil

import hashlib


class BCURStringFormatError(RuntimeError):
    pass


def bcur_encode(data):
    """Returns bcur encoded string and checksum"""
    cbor = cbor_encode(data)
    enc = bc32encode(cbor)
    h = hashlib.sha256(cbor).digest()
    enc_hash = bc32encode(h)
    return enc, enc_hash


def bcur_decode(data, checksum=None):
    """Returns decoded data, verifies checksum if provided"""
    cbor = bc32decode(data)
    if checksum is not None:
        h = bc32decode(checksum)
        calculated_digest = hashlib.sha256(cbor).digest()
        if h != calculated_digest:
            raise ValueError(f"Calculated digest {calculated_digest} != {h}")
    return cbor_decode(cbor)


def _parse_bcur_helper(bcur_string):
    """
    This parses a bcur string and returns the following (or raises an error):

        payload, checksum, x, y

    Notes:
    - Works for both BCURSingle and BCURMulti.
    - All entries may be empty except for payload.
    - Checksums are not validated here, as checksum validation is different for single vs multi.
    """

    if type(bcur_string) is not str:
        raise BCURStringFormatError(
            f"{bcur_string} is of type {type(bcur_string)}, not a string"
        )

    string = bcur_string.lower().strip()

    if not string.startswith("ur:bytes/"):
        raise BCURStringFormatError(f"String {string} doesn't start with ur:bytes/")

    bcur_parts = string.split("/")
    if len(bcur_parts) == 2:
        # Non-animated QR code (just 1 qr, doesn't display 1of1 nor checksum)
        _, payload = bcur_parts
        checksum, x_int, y_int = None, 1, 1
    elif len(bcur_parts) == 3:
        # Non-animated QR code (just 1 qr, doesn't display 1of1 but does have checksum)
        _, checksum, payload = bcur_parts
        x_int, y_int = 1, 1
    elif len(bcur_parts) == 4:
        # Animated QR code
        _, xofy, checksum, payload = bcur_parts

        xofy_parts = xofy.split("of")
        if len(xofy_parts) != 2:
            raise BCURStringFormatError(f"x-of-y section malformed: {xofy_parts}")

        if not is_intable(xofy_parts[0]) or not is_intable(xofy_parts[1]):
            raise BCURStringFormatError(
                f"x and y (in x-of-y) must both be integers: {xofy_parts}"
            )

        x_int = int(xofy_parts[0])
        y_int = int(xofy_parts[1])

        if x_int > y_int:
            raise BCURStringFormatError("x must be >= y (in x-of-y): {xofy_parts}")

    else:
        raise BCURStringFormatError(f"{string} doesn't have 2-4 slashes")

    if checksum:
        if len(checksum) != 58:
            raise BCURStringFormatError("Checksum must be 58 chars")
        if not uses_only_bech32_chars(checksum):
            raise BCURStringFormatError(
                f"checksum can only contain bech32 characters: {checksum}"
            )

    if not uses_only_bech32_chars(payload):
        raise BCURStringFormatError(
            f"Payload can only contain bech32 characters: {payload}"
        )

    return payload, checksum, x_int, y_int


class BCURSingle:
    def __init__(self, text_b64, encoded=None, checksum=None):
        binary_b64 = a2b_base64(text_b64)
        enc, enc_hash = bcur_encode(data=binary_b64)
        if encoded and encoded != enc:
            raise ValueError(f"Calculated encoding {enc} != {encoded}")

        if checksum and checksum != enc_hash:
            raise ValueError(f"Calculated checksum {enc_hash} != {checksum}")

        self.text_b64 = text_b64
        self.encoded = enc
        self.enc_hash = enc_hash

    def __repr__(self):
        return self.encode()

    def encode(self, use_checksum=True):
        # Single QR, no x-of-y
        if use_checksum:
            return f"ur:bytes/{self.enc_hash}/{self.encoded}"
        else:
            return f"ur:bytes/{self.encoded}"

    @classmethod
    def parse(cls, to_parse):
        """Parses (decodes) a BCURSingle from a single BCUR string"""

        payload, checksum, x, y = _parse_bcur_helper(bcur_string=to_parse)

        if x != 1 or y != 1:
            raise BCURStringFormatError(
                f"BCURSingle must have x=1 and y=1, instead got x={x} and y={y}"
            )

        # will throw an error if checksum is incorrect
        enc = bcur_decode(data=payload, checksum=checksum)
        return cls(
            text_b64=b2a_base64(enc).strip().decode(),
            encoded=payload,
            checksum=checksum,
        )


class BCURMulti:
    def __init__(self, text_b64, encoded=None, checksum=None):
        binary_b64 = a2b_base64(text_b64)
        enc, enc_hash = bcur_encode(data=binary_b64)
        if encoded and encoded != enc:
            raise ValueError(f"Calculated encoding {enc} != {encoded}")

        if checksum and checksum != enc_hash:
            raise ValueError(f"Calculated checksum {enc_hash} != {checksum}")

        self.checksum = checksum
        self.encoded = enc
        self.text_b64 = text_b64
        self.enc_hash = enc_hash

    def __repr__(self):
        return f"bcur: {self.checksum}\n{self.text_b64}\n"

    def encode(self, max_size_per_chunk=300, animate=True):
        """
        Take some base64 text (i.e. a PSBT string) and encode it into multiple QR codes using Blockchain Commons Uniform Resources.

        If animate=False, then max_size_per_chunk is ignored and this returns a 1of1 with checksum.

        Use parse() to return a BCURMulti object from this encoded result.

        This algorithm makes all the chunks of about equal length.
        This makes sure that the last chunk is not (too) different in size which is visually noticeable when animation occurs
        Inspired by this JS implementation:
        https://github.com/cryptoadvance/specter-desktop/blob/da35e7d88072475746077432710c77f799017eb0/src/cryptoadvance/specter/templates/includes/qr-code.html
        """

        if animate is False:
            number_of_chunks = 1
        else:
            number_of_chunks = ceil(len(self.encoded) / max_size_per_chunk)

        chunk_length = ceil(len(self.encoded) / number_of_chunks)

        # For number_of_chunks == 1 (with no checksum) use BCURSingle

        resulting_chunks = []
        for cnt in range(number_of_chunks):
            start_idx = cnt * chunk_length
            finish_idx = (cnt + 1) * chunk_length
            resulting_chunks.append(
                f"ur:bytes/{cnt+1}of{number_of_chunks}/{self.enc_hash}/{self.encoded[start_idx:finish_idx]}"
            )

        return resulting_chunks

    @classmethod
    def parse(cls, to_parse):
        """Parses a BCURMulti from a list of BCUR strings"""
        if type(to_parse) not in (list, tuple):
            raise BCURStringFormatError(
                f"{to_parse} is of type {type(to_parse)}, not a list/tuple"
            )

        payloads = []
        global_checksum, global_y = "", 0
        for cnt, bcur_string in enumerate(to_parse):
            entry_payload, entry_checksum, entry_x, entry_y = _parse_bcur_helper(
                bcur_string=bcur_string
            )
            if cnt + 1 != entry_x:
                raise ValueError(
                    f"BCUR strings not in order: got {entry_x} and was expecting {cnt+1}"
                )

            # Initialize checksum and y (as in x-of-y) on first loop
            if cnt == 0:
                global_checksum = entry_checksum
                global_y = entry_y

            elif entry_checksum != global_checksum:
                raise ValueError(
                    f"Entry {bcur_string} has checksum {entry_checksum} but we're expecting {global_checksum}"
                )
            elif entry_y != global_y:
                raise ValueError(
                    f"Entry {bcur_string} wants {entry_y} parts but we're expecting {global_y} parts"
                )
            # All checks pass
            payloads.append(entry_payload)

        # will throw an error if checksum is incorrect
        enc = bcur_decode(data="".join(payloads), checksum=global_checksum)

        return cls(text_b64=b2a_base64(enc).strip().decode(), checksum=global_checksum)



# TODO: specify what specifically to import
from .bcur import *
from .bech32 import *
from .blinding import *
from .block import *
from .bloomfilter import *
from .descriptor import *
from .ecc import *
from .hd import *
from .helper import *
from .merkleblock import *
from .mnemonic import *
from .network import *
from .op import *
from .pbkdf2 import *
from .psbt import *
from .psbt_helper import *
from .script import *
from .shamir import *
from .tx import *
from .witness import *



from buidl.helper import (
    encode_varint,
    encode_varstr,
    read_varint,
    read_varstr,
)


class Witness:
    def __init__(self, items=None):
        self.items = items or []

    def __repr__(self):
        result = ""
        for item in self.items:
            result += "{} ".format(item.hex())
        return result

    @classmethod
    def parse(cls, s):
        num_items = read_varint(s)
        items = []
        for _ in range(num_items):
            items.append(read_varstr(s))
        return cls(items)

    def __len__(self):
        return len(self.items)

    def serialize(self):
        result = encode_varint(len(self))
        for item in self.items:
            if len(item) == 1:
                result += item
            else:
                result += encode_varstr(item)
        return result



from unittest import TestCase

from buidl.mnemonic import secure_mnemonic
from buidl.hd import HDPrivateKey


class MnemonicTest(TestCase):
    def test_secure_mnemonic_bits(self):
        tests = (
            # num_bits, num_words
            (128, 12),
            (160, 15),
            (192, 18),
            (224, 21),
            (256, 24),
        )

        for num_bits, num_words in tests:
            mnemonic = secure_mnemonic(num_bits=num_bits)
            self.assertEqual(num_words, len(mnemonic.split(" ")))
            # This is inherently non-deterministic, so we can't check the specific output
            HDPrivateKey.from_mnemonic(mnemonic, network="testnet")

        for invalid_num_bits in (-1, 1, 127, 129, 257, "notanint"):
            with self.assertRaises(ValueError):
                secure_mnemonic(num_bits=invalid_num_bits)

    def test_secure_mnemonic_extra_entropy(self):
        tests = (
            # num_bits, num_words, extra_entropy
            (128, 12, 0),
            (160, 15, 1),
            (192, 18, 2 ** 128),
            (224, 21, 2 ** 256),
            (256, 24, 2 ** 512),
        )

        for num_bits, num_words, extra_entropy in tests:
            mnemonic = secure_mnemonic(num_bits=num_bits, extra_entropy=extra_entropy)
            self.assertEqual(num_words, len(mnemonic.split(" ")))
            # This is inherently non-deterministic, so we can't check the specific output
            HDPrivateKey.from_mnemonic(mnemonic, network="testnet")

        with self.assertRaises(TypeError):
            secure_mnemonic(extra_entropy="not an int")
        with self.assertRaises(ValueError):
            secure_mnemonic(extra_entropy=-1)



from unittest import TestCase
import unittest  # for skipunless

from io import BytesIO
from os import getenv

from buidl.block import Block
from buidl.helper import decode_base58, decode_gcs
from buidl.network import (
    BASIC_FILTER_TYPE,
    CFCheckPointMessage,
    CFHeadersMessage,
    CFilterMessage,
    FILTERED_BLOCK_DATA_TYPE,
    GetDataMessage,
    GetHeadersMessage,
    GetCFCheckPointMessage,
    GetCFHeadersMessage,
    GetCFiltersMessage,
    HeadersMessage,
    NetworkEnvelope,
    SimpleNode,
    VersionMessage,
)


class NetworkEnvelopeTest(TestCase):
    def test_parse(self):
        msg = bytes.fromhex("f9beb4d976657261636b000000000000000000005df6e0e2")
        stream = BytesIO(msg)
        envelope = NetworkEnvelope.parse(stream)
        self.assertEqual(envelope.command, b"verack")
        self.assertEqual(envelope.payload, b"")
        msg = bytes.fromhex(
            "f9beb4d976657273696f6e0000000000650000005f1a69d2721101000100000000000000bc8f5e5400000000010000000000000000000000000000000000ffffc61b6409208d010000000000000000000000000000000000ffffcb0071c0208d128035cbc97953f80f2f5361746f7368693a302e392e332fcf05050001"
        )
        stream = BytesIO(msg)
        envelope = NetworkEnvelope.parse(stream)
        self.assertEqual(envelope.command, b"version")
        self.assertEqual(envelope.payload, msg[24:])

    def test_serialize(self):
        msg = bytes.fromhex("f9beb4d976657261636b000000000000000000005df6e0e2")
        stream = BytesIO(msg)
        envelope = NetworkEnvelope.parse(stream)
        self.assertEqual(envelope.serialize(), msg)
        msg = bytes.fromhex(
            "f9beb4d976657273696f6e0000000000650000005f1a69d2721101000100000000000000bc8f5e5400000000010000000000000000000000000000000000ffffc61b6409208d010000000000000000000000000000000000ffffcb0071c0208d128035cbc97953f80f2f5361746f7368693a302e392e332fcf05050001"
        )
        stream = BytesIO(msg)
        envelope = NetworkEnvelope.parse(stream)
        self.assertEqual(envelope.serialize(), msg)


class VersionMessageTest(TestCase):
    def test_serialize(self):
        v = VersionMessage(timestamp=0, nonce=b"\x00" * 8)
        self.assertEqual(
            v.serialize().hex(),
            "7f11010000000000000000000000000000000000000000000000000000000000000000000000ffff000000008d20000000000000000000000000000000000000ffff000000008d2000000000000000001b2f70726f6772616d6d696e67626c6f636b636861696e3a302e312f0000000001",
        )


class GetHeadersMessageTest(TestCase):
    def test_serialize(self):
        block_hex = "0000000000000000001237f46acddf58578a37e213d2a6edc4884a2fcad05ba3"
        gh = GetHeadersMessage(start_block=bytes.fromhex(block_hex))
        self.assertEqual(
            gh.serialize().hex(),
            "7f11010001a35bd0ca2f4a88c4eda6d213e2378a5758dfcd6af437120000000000000000000000000000000000000000000000000000000000000000000000000000000000",
        )


class HeadersMessageTest(TestCase):
    def test_parse(self):
        hex_msg = "0200000020df3b053dc46f162a9b00c7f0d5124e2676d47bbe7c5d0793a500000000000000ef445fef2ed495c275892206ca533e7411907971013ab83e3b47bd0d692d14d4dc7c835b67d8001ac157e670000000002030eb2540c41025690160a1014c577061596e32e426b712c7ca00000000000000768b89f07044e6130ead292a3f51951adbd2202df447d98789339937fd006bd44880835b67d8001ade09204600"
        stream = BytesIO(bytes.fromhex(hex_msg))
        headers = HeadersMessage.parse(stream)
        self.assertEqual(len(headers.headers), 2)
        for b in headers.headers:
            self.assertEqual(b.__class__, Block)


class GetDataMessageTest(TestCase):
    def test_serialize(self):
        hex_msg = "020300000030eb2540c41025690160a1014c577061596e32e426b712c7ca00000000000000030000001049847939585b0652fba793661c361223446b6fc41089b8be00000000000000"
        get_data = GetDataMessage()
        block1 = bytes.fromhex(
            "00000000000000cac712b726e4326e596170574c01a16001692510c44025eb30"
        )
        get_data.add_data(FILTERED_BLOCK_DATA_TYPE, block1)
        block2 = bytes.fromhex(
            "00000000000000beb88910c46f6b442312361c6693a7fb52065b583979844910"
        )
        get_data.add_data(FILTERED_BLOCK_DATA_TYPE, block2)
        self.assertEqual(get_data.serialize().hex(), hex_msg)


@unittest.skipUnless(
    getenv("INCLUDE_NETWORK_TESTS"),
    reason="Requires network connection, so may not be unreliable",
)
class SimpleNodeTest(TestCase):
    def test_handshake(self):
        node = SimpleNode("testnet.programmingbitcoin.com", network="testnet")
        node.handshake()

    def test_handshake_signet(self):
        node = SimpleNode("signet.programmingbitcoin.com", network="signet")
        node.handshake()

    def test_get_filtered_txs(self):
        from buidl.bloomfilter import BloomFilter

        bf = BloomFilter(30, 5, 90210)
        h160 = decode_base58("mseRGXB89UTFVkWJhTRTzzZ9Ujj4ZPbGK5")
        bf.add(h160)
        node = SimpleNode("testnet.programmingbitcoin.com", network="testnet")
        node.handshake()
        node.send(bf.filterload())
        block_hash = bytes.fromhex(
            "00000000000377db7fde98411876c53e318a395af7304de298fd47b7c549d125"
        )
        txs = node.get_filtered_txs([block_hash])
        self.assertEqual(
            txs[0].id(),
            "0c024b9d3aa2ae8faae96603b8d40c88df2fc6bf50b3f446295206f70f3cf6ad",
        )
        self.assertEqual(
            txs[1].id(),
            "0886537e27969a12478e0d33707bf6b9fe4fdaec8d5d471b5304453b04135e7e",
        )
        self.assertEqual(
            txs[2].id(),
            "23d4effc88b80fb7dbcc2e6a0b0af9821c6fe3bb4c8dc3b61bcab7c45f0f6888",
        )


class CFilterTest(TestCase):
    def test_cfilter(self):
        stop_hash = bytes.fromhex(
            "000000006f27ddfe1dd680044a34548f41bed47eba9e6f0b310da21423bc5f33"
        )
        getcfilters = GetCFiltersMessage(stop_hash=stop_hash)
        expected = b"\x00\x01\x00\x00\x00" + stop_hash[::-1]
        self.assertEqual(getcfilters.serialize(), expected)
        expected = (
            b"\x00" + stop_hash[::-1] + b"\x09" + bytes.fromhex("0385acb4f0fe889ef0")
        )
        cfilter = CFilterMessage.parse(BytesIO(expected))
        self.assertEqual(cfilter.filter_type, 0)
        self.assertEqual(cfilter.block_hash, stop_hash)
        self.assertEqual(cfilter.hashes, {1341840, 1483084, 570774})
        self.assertEqual(cfilter.hash(b"\x00"), 1322199)
        included = bytes.fromhex(
            "002027a5000c7917f785d8fc6e5a55adfca8717ecb973ebb7743849ff956d896a7ed"
        )
        self.assertTrue([included] in cfilter)
        self.assertFalse([b"\x00"] in cfilter)
        with self.assertRaises(RuntimeError):
            GetCFiltersMessage()

    def test_cfilter_without_network(self):
        # Example from Trezor Blog Post (https://blog.trezor.io/bip158-compact-block-filters-9b813b07a878)
        block_hash_hex = (
            "000000000000015d6077a411a8f5cc95caf775ccf11c54e27df75ce58d187313"
        )
        filter_hex = "09027acea61b6cc3fb33f5d52f7d088a6b2f75d234e89ca800"
        key = bytes.fromhex(block_hash_hex)[::-1][:16]
        filter_bytes = bytes.fromhex(filter_hex)
        cfilter = CFilterMessage(
            filter_type=BASIC_FILTER_TYPE,
            block_hash=bytes.fromhex(block_hash_hex),
            filter_bytes=filter_bytes,
            hashes=decode_gcs(key, filter_bytes),
        )
        for script, want in (
            ("76a9143ebc40e411ed3c76f86711507ab952300890397288ac", True),
            ("76a914c01a7ca16b47be50cbdbc60724f701d52d75156688ac", True),
            ("76a914000000000000000000000000000000000000000088ac", False),  # made up
        ):
            self.assertEqual(bytes.fromhex(script) in cfilter, want)


class CFHeaderTest(TestCase):
    def test_cfheader(self):
        stop_hash = bytes.fromhex(
            "000000006f27ddfe1dd680044a34548f41bed47eba9e6f0b310da21423bc5f33"
        )
        getcfheaders = GetCFHeadersMessage(stop_hash=stop_hash)
        self.assertEqual(
            getcfheaders.serialize(), b"\x00\x00\x00\x00\x00" + stop_hash[::-1]
        )
        hash2 = b"\x00" * 32
        stream = BytesIO(
            bytes.fromhex(
                "00335fbc2314a20d310b6f9eba7ed4be418f54344a0480d61dfedd276f000000000000000000000000000000000000000000000000000000000000000000000000010000000000000000000000000000000000000000000000000000000000000000"
            )
        )
        cfheaders = CFHeadersMessage.parse(stream)
        self.assertEqual(cfheaders.filter_type, 0)
        self.assertEqual(cfheaders.stop_hash, stop_hash)
        self.assertEqual(cfheaders.previous_filter_header, hash2)
        self.assertEqual(cfheaders.filter_hashes, [hash2])
        with self.assertRaises(RuntimeError):
            GetCFHeadersMessage()


class CFCheckPointTest(TestCase):
    def test_cfcheckpoint(self):
        stop_hash = bytes.fromhex(
            "000000006f27ddfe1dd680044a34548f41bed47eba9e6f0b310da21423bc5f33"
        )
        getcfcheckpoints = GetCFCheckPointMessage(stop_hash=stop_hash)
        self.assertEqual(getcfcheckpoints.serialize(), b"\x00" + stop_hash[::-1])
        hash2 = b"\x00" * 32
        stream = BytesIO(
            bytes.fromhex(
                "00335fbc2314a20d310b6f9eba7ed4be418f54344a0480d61dfedd276f00000000010000000000000000000000000000000000000000000000000000000000000000000000"
            )
        )
        cfcheckpoints = CFCheckPointMessage.parse(stream)
        self.assertEqual(cfcheckpoints.filter_type, 0)
        self.assertEqual(cfcheckpoints.stop_hash, stop_hash)
        self.assertEqual(cfcheckpoints.filter_headers, [hash2])
        with self.assertRaises(RuntimeError):
            GetCFCheckPointMessage()



from unittest import TestCase

from buidl.bcur import (
    bcur_encode,
    bcur_decode,
    BCURSingle,
    BCURMulti,
    _parse_bcur_helper,
    BCURStringFormatError,
)

from binascii import a2b_base64
from base64 import b64encode


class SpecterDesktopTest(TestCase):
    def test_specter_desktop(self):
        """
        Test case from specter-desktop:
        https://github.com/cryptoadvance/specter-desktop/blob/0a483316d0d2e83cb5a532a0cbbcd82a885587db/src/cryptoadvance/specter/util/bcur.py
        """

        psbt_b64 = "cHNidP8BAHEBAAAAAfPQ5Rpeu5nH0TImK4Sbu9lxIOGEynRadywPxaPyhnTwAAAAAAD/////AkoRAAAAAAAAFgAUFCYoQzGSRmYVAuZNuXF0OrPg9jWIEwAAAAAAABYAFOZMlwM1sZGLivwOcOh77amAlvD5AAAAAAABAR+tKAAAAAAAABYAFM4u9V5WG+Fe9l3MefmYEX4ULWAWIgYDA+jO+oOuN37ABK67BA/+SuuR/57c7OkyfyR7hR34FDsYccBxUlQAAIAAAACAAAAAgAAAAAAFAAAAACICApJMZBvzWiavLN7nievKQoylwPoffLkXZUIgGHF4HgwaGHHAcVJUAACAAAAAgAAAAIABAAAACwAAAAAA"
        raw = a2b_base64(psbt_b64)
        enc, enc_hash = bcur_encode(raw)
        dec = bcur_decode(enc, enc_hash)
        self.assertEqual(dec, raw)
        testres = [
            "tyq3wurnvf607qgqwyqsqqqqq8eapeg6t6aen373xgnzhpymh0vhzg8psn98gknh9s8utgljse60qqqqqqqqpllllllsyjs3qqqqqqqqqqtqq9q5yc5yxvvjgenp2qhxfkuhzap6k0s0vdvgzvqqqqqqqqqpvqq5uexfwqe4kxgchzhupecws7ld4xqfdu8eqqqqqqqq",
            "qyq3ltfgqqqqqqqqqqtqq9xw9m64u4smu900vhwv08uesyt7zskkq93zqcps86xwl2p6udm7cqz2awcyplly46u3l70dem8fxfljg7u9rhupgwccw8q8z5j5qqqgqqqqqzqqqqqqsqqqqqqqq5qqqqqqygpq9yjvvsdlxk3x4ukdaeufa09y9r99crap7l9ezaj5ygqc",
            "w9upurq6rpcuqu2j2sqqpqqqqqqgqqqqqzqqzqqqqq9sqqqqqqqqmkdau4",
        ]
        testhash = "hlwjxjx550k4nnfdl5py2tn3vnh6g60slnw5dmld6ktrkkz200as49spg5"
        self.assertEqual(enc_hash, testhash)
        self.assertEqual(enc, "".join(testres))

        # Test by instantiating a BCURSingle object
        bcur_single_obj = BCURSingle(text_b64=psbt_b64)
        self.assertEqual(bcur_single_obj.enc_hash, enc_hash)
        self.assertEqual(bcur_single_obj.encoded, enc)
        self.assertEqual(bcur_single_obj.text_b64, psbt_b64)
        self.assertEqual(
            bcur_single_obj.encode(use_checksum=True), f"ur:bytes/{enc_hash}/{enc}"
        )
        self.assertEqual(bcur_single_obj.encode(use_checksum=False), f"ur:bytes/{enc}")

        # Re-instantiate this object from the encoded version
        bcur_single_obj_2 = BCURSingle.parse(to_parse=f"ur:bytes/{enc}")
        self.assertEqual(bcur_single_obj.encode(), bcur_single_obj_2.encode())

        # Make it into a multi:
        expected_multi = []
        for cnt, res in enumerate(testres):
            expected_multi.append(f"ur:bytes/{cnt+1}of{len(testres)}/{testhash}/{res}")
        bcur_multi_obj = BCURMulti.parse(to_parse=expected_multi)
        self.assertEqual(bcur_multi_obj.text_b64, psbt_b64)


class BCURMultiTest(TestCase):
    def test_single_frame_qrgif(self):
        chunks = BCURMulti(text_b64=b64encode(b"foo")).encode(animate=False)
        self.assertEqual(len(chunks), 1)
        self.assertEqual(
            chunks[0],
            "ur:bytes/1of1/j7snj9l0tttmp4c0d9d9mdz0frkac8s6fz4cn8erca3nxz0cnjuq7fv7lv/gdnx7mc0p7099",
        )

    def test_bcur_multi_encoding_decoding(self):
        psbt_b64 = "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"

        chunks_calculated = BCURMulti(text_b64=psbt_b64).encode(
            max_size_per_chunk=300, animate=True
        )
        checksum_expected = "qud9jpcv0af7refxxwu7dvhqshmct65hdzjkpu9q85hw7vkmvshqs2snsx"
        chunks_expected = [
            "tyrukurnvf607qgqe5pqqqqqqsdasdzvck45x4ktrm5jg8nqvuxazw4zhrupuz5y63c9x32c8z7l7qqqqqqqpl0llll6uhht2jztjuekypk9gdsu307alw37afyjd2dzmzhue0g30634fyqqqqqqqq8allll7xd6tl5y20ta4k3wd9382m0kmfw4y5fp35c74rgjszstk72vqp08qqqqqqqqlhlllluhse9w6m4zy0h0t3c6slp6m6qxld8283t0fj2h3qrycf20g2dpfyqsqqqqqr7lllllq8y3",
            "2qqqqqqqqqqkqq2dvue6qn8ep555djamy6vxr9kfk8sx3vssqqqqqqqqzqfts5yqqqqqqqqqqgsqyrzmapaunvc7smex3stl7ygs94tyy3e5mwue7jm6yknxheu9wuaquqg93dgjzq3at9hf286k8gsv23t7a8wuezqcxmnrylf2qjgrgkrjz9e684sktsss9jfyeuqcr7zwmf8h3xcdvg2c92vz64dxkzr49tswx9vhx2y5myqayypd7pseq3autesaqjnd9axjhdjajygth8es7ulpcpnwa8x9",
            "u3ctfqppq0f5svtqyhxl6jz8atxdk3lxrk4elgx9jduwgzdl4m5zqmj99eytg49wygrqy02ed6237436yrz52lhfmhxgsxpkuce862syjq69sus3ww3av9jurnmaqsysxqqqpqqpqqqgqqqqqzqqyqqqsqqsqqqqqgqqqqpzqcpvjfx0qxqlsnk6faufkrtzzkp2nqk4tf4ssaf2ur33t9ej39xeq8gu8ffttnfsqqqgqqgqqzqqqqqqsqpqqqyqqyqqqqqzqqqqqgsxqt0svxgy00z7v8gy5mf0",
            "f54mvhv3zzae7v8h8cwqvmhfe30ywz6gq8qjnq8w6vqqqzqqzqqqsqqqqqyqqgqqpqqpqqqqqqsqqqqzypsr6dyrzcp9eh75s3l2end50esa4w06p3vn0rjqn0awaqsxu3fwfz6pe37svj9rqqqqsqqsqqyqqqqqpqqzqqqgqqgqqqqqyqqqqqqqzqftxvpsqqqqqqqqqgsqypxkj40yu9s24vm9nxfq7qav6j6c3yepwc9rf0jscx7v38tafrz2zqg93dgjzq3k5m85y4xg9y9pdrk2kjhwwugp",
            "356ha2r32jjmtl4fakd6acjctcssx40vzqqu9383mn3da9qtat9aed7hw3s5q2q6j2psqz4yd5j36333yypcx0tw03qjzxq0k7gcpdu2q4e6647znxp97x85na559jeck6lsyw3pqw57xswr9ky8qurpz4zreutrhlpa9ksv4pg452ducxjsp3julv3mk49wygrqyd4xeap9fjpfpgtgaj454mnhzqvdx4l2su255k6la20dnwhwykz7rnraqey2xqqqpqqpqqqgqqqqqzqqyqqqsqqsqqqqqqqq",
            "qqpzqcp4tmqsq8pvfuwuut0fgzl2e0wt04m5v9qzsx5jsvqq4frdy5w5vvgu8ffttnfsqqqgqqgqqzqqqqqqsqpqqqyqqyqqqqqqqqqqqgsxqwpn6mnugys3srahjxqt0zs9wwk40s5esf033ayld9pvkw9khupr58qjnq8w6vqqqzqqzqqqsqqqqqyqqgqqpqqpqqqqqqqqqqqzypsr4835rsed3pc8qcg4gs70zcalc0fd5r9g29dzn0xp55qvvh8mywa3ea7sgzgrqqqqsqqsqqyqqqqqpqqz",
            "qqqgqqgqqqqqqqqqqqqqzqft6qrsqqqqqqqqqgsqyzqy6gtmryq99fyjde5t85de57w4umkv3j8uc0m3q02pptqaffpxjqg93dgjzq4n6tcd2vnpsc97224c3hvrqenhttjklaclcflytu95mu3qqg7kzgss9hfmns4pg59ud6pnwumygpfpzlex6lxvga2rtq6xh3wyzpkat698yypsfq9w7rfvl7qefnmynmqc2p7vwnqgpvalusrjk27k50a579sxy6ppqwwfasmz9l0lvsx5xa0kunyla7uk",
            "cczppesja3qz3vgvr260uw8sx49wygrq9v7j7r2nycvxp0jj4wydmqcxva66u4hlw87z0ezlpdxlygqz84sjrnraqey2xqqqpqqpqqqgqqqqqzqqyqqqsqqqqqqqq5qqqqpzqcpd6wuu9g29p0rwsvmhxezq2ggh7fkhenz82s6cx34ut3qsdh273fcu7lgypypsqqqgqqgqqzqqqqqqsqpqqqyqqqqqqqq9qqqqqgsxqvzgpths6t8lsx2v7ey7cxzs0nr5czqt80lyqu4jh44rld83vp3xs8p6",
            "226u6vqqqzqqzqqqsqqqqqyqqgqqpqqqqqqqqpgqqqqzypsrnj0vxc30mlmyp4phtahye8l0h9kxqsgwvyhvgq5tzrq6knlr3up3cy5cpmknqqqqsqqsqqyqqqqqpqqzqqqgqqqqqqqq2qqqqqqqzqftaqpsqqqqqqqqqgsqyp7l65xe5xlkjqtedv8qxykzhjdg00kayjzxc0rnm2kp3kxcmkz6xqg93dgjzq3ke5mu4gfxzhp8vnzchxhptpjztzwm3fjwvd85m9v864plvglwr5ss93q68nq6",
            "5f33gsgqxgw760vljprc5shklwhkklr7xulh4lqnpchdyype2cej0zs0wuzkmmjyzzhdj6pehskmdfyvtn643usf32d7v4hquzppqwvhggjan6vcftzxwjwswuctvwwfeg5kehk48qujp88amtsy3004j49wygrqydkdxl92zfs4cfmyck9e4c2cvsjcnku2vnnrfaxetp74g0mz8msarnmaqsysxqqqpqqpqqqgqqqqqzqqyqqqsqqqqqqqqvqqqqpzqcpvgx3ucx4zvv2yzqpjrhknm8usg79y",
            "9ahm4a4hcl3h8aa0cycw9mguz2vqamfsqqqgqqgqqzqqqqqqsqpqqqyqqqqqqqqrqqqqqgsxqw2kxvnc5rmhq4k7u3qs4mvkswdu9km2frzu74v0yzv2n0n9dc8qs8x86pjg5vqqqzqqzqqqsqqqqqyqqgqqpqqqqqqqqqcqqqqzypsrn96zyhv7nxz2c3n5n5rhxzmrnjw299kda4fc8ysfelw6upytmav3cwjjkhxnqqqqsqqsqqyqqqqqpqqzqqqgqqqqqqqqxqqqqqqqqw35949",
        ]
        for cnt, chunk in enumerate(chunks_calculated):
            self.assertEqual(
                f"ur:bytes/{cnt+1}of{len(chunks_calculated)}/{checksum_expected}/{chunks_expected[cnt]}",
                chunk,
            )

        bcur_multi_obj = BCURMulti.parse(to_parse=chunks_calculated)
        self.assertEqual(bcur_multi_obj.text_b64, psbt_b64)
        self.assertEqual(bcur_multi_obj.encoded, "".join(chunks_expected))
        self.assertEqual(bcur_multi_obj.enc_hash, checksum_expected)


class ParseTest(TestCase):
    def setUp(self):
        self.GOOD_CHECKSUM = (
            "qud9jpcv0af7refxxwu7dvhqshmct65hdzjkpu9q85hw7vkmvshqs2snsx"
        )
        self.GOOD_ENCODED = "tyrukurnvf607qgqe5pqqqqqqsdasdzvck45x4ktrm5jg8nqvuxazw4zhrupuz5y63c9x32c8z7l7qqqqqqqpl0llll6uhht2jztjuekypk9gdsu307alw37afyjd2dzmzhue0g30634fyqqqqqqqq8allll7xd6tl5y20ta4k3wd9382m0kmfw4y5fp35c74rgjszstk72vqp08qqqqqqqqlhlllluhse9w6m4zy0h0t3c6slp6m6qxld8283t0fj2h3qrycf20g2dpfyqsqqqqqr7lllllq8y3"

    def test_valid_bcur_parse_single_yeschecksum(self):
        bcur_string = f"ur:bytes/{self.GOOD_CHECKSUM}/{self.GOOD_ENCODED}"
        payload_parsed, checksum_parsed, x, y = _parse_bcur_helper(bcur_string)
        self.assertEqual(x, 1)
        self.assertEqual(y, 1)
        self.assertEqual(checksum_parsed, self.GOOD_CHECKSUM)
        self.assertEqual(payload_parsed, self.GOOD_ENCODED)

    def test_valid_bcur_parse_single_nochecksum(self):
        bcur_string = f"ur:bytes/{self.GOOD_ENCODED}"
        payload_parsed, checksum_parsed, x, y = _parse_bcur_helper(bcur_string)
        self.assertEqual(x, 1)
        self.assertEqual(y, 1)
        self.assertEqual(checksum_parsed, None)
        self.assertEqual(payload_parsed, self.GOOD_ENCODED)

    def test_valid_bcur_parse_multi(self):
        bcur_string = f"ur:bytes/2of8/{self.GOOD_CHECKSUM}/{self.GOOD_ENCODED}"
        payload_parsed, checksum_parsed, x, y = _parse_bcur_helper(bcur_string)
        self.assertEqual(x, 2)
        self.assertEqual(y, 8)
        self.assertEqual(checksum_parsed, self.GOOD_CHECKSUM)
        self.assertEqual(payload_parsed, self.GOOD_ENCODED)

    def test_invalid_bcur_parse(self):
        with self.assertRaises(BCURStringFormatError):
            _parse_bcur_helper("fail")

        with self.assertRaises(BCURStringFormatError):
            _parse_bcur_helper("ur:bytes/1/2/3")

        with self.assertRaises(BCURStringFormatError):
            _parse_bcur_helper("ur:bytes/foo")

        with self.assertRaises(BCURStringFormatError):
            _parse_bcur_helper("ur:bytes/gd56dxsyew2w5/foo")


class BCURSingleTest(TestCase):
    def test_simple_bcur(self):
        """ 1of1 with and without checksum """

        # Basic test
        bcur_single_obj = BCURSingle(text_b64="aaaa")
        self.assertEqual(
            bcur_single_obj.enc_hash,
            "ysypyck5etagxt08hzn6vcnwam3lgupp0uhcs7n8pg0wmen32p3qate5eg",
        )
        self.assertEqual(bcur_single_obj.encoded, "gd56dxsyew2w5")

        # Render with and without checksum
        bcur_single_checksum = "ur:bytes/ysypyck5etagxt08hzn6vcnwam3lgupp0uhcs7n8pg0wmen32p3qate5eg/gd56dxsyew2w5"
        bcur_single_nochecksum = "ur:bytes/gd56dxsyew2w5"
        self.assertEqual(
            bcur_single_obj.encode(use_checksum=True), bcur_single_checksum
        )
        self.assertEqual(
            bcur_single_obj.encode(use_checksum=False), bcur_single_nochecksum
        )

        # Reconstruct from bcur string with checksum
        bcur_single_reconstructed_v1 = BCURSingle.parse(to_parse=bcur_single_checksum)
        self.assertEqual(
            bcur_single_reconstructed_v1.encode(use_checksum=True), bcur_single_checksum
        )
        self.assertEqual(
            bcur_single_reconstructed_v1.encode(use_checksum=False),
            bcur_single_nochecksum,
        )

        # Reconstruct from bcur string without checksum
        bcur_single_reconstructed_v2 = BCURSingle.parse(to_parse=bcur_single_nochecksum)
        self.assertEqual(
            bcur_single_reconstructed_v2.encode(use_checksum=True), bcur_single_checksum
        )
        self.assertEqual(
            bcur_single_reconstructed_v2.encode(use_checksum=False),
            bcur_single_nochecksum,
        )


class BCURDecodeTest(TestCase):
    def test_bcur_encode_decode(self):
        # See that encode/decode works
        b64_text = b64encode(b"foo")
        raw = a2b_base64(b64_text)
        enc, enc_hash = bcur_encode(raw)
        dec = bcur_decode(enc, checksum=enc_hash)
        self.assertEqual(dec, raw)

        # Confirm bad hash raises an error
        with self.assertRaises(ValueError):
            bcur_decode(enc, checksum=enc_hash + "fail")



import re

from io import BytesIO

from buidl.helper import int_to_big_endian

BECH32_ALPHABET = "qpzry9x8gf2tvdw0s3jn54khce6mua7l"
GEN = [0x3B6A57B2, 0x26508E6D, 0x1EA119FA, 0x3D4233DD, 0x2A1462B3]

BECH32_CHARS_RE = re.compile("^[qpzry9x8gf2tvdw0s3jn54khce6mua7l]*$")


def uses_only_bech32_chars(string):
    return bool(BECH32_CHARS_RE.match(string.lower()))


# next four functions are straight from BIP0173:
# https://github.com/bitcoin/bips/blob/master/bip-0173.mediawiki
def bech32_polymod(values):
    chk = 1
    for v in values:
        b = chk >> 25
        chk = (chk & 0x1FFFFFF) << 5 ^ v
        for i in range(5):
            chk ^= GEN[i] if ((b >> i) & 1) else 0
    return chk


def bech32_hrp_expand(s):
    b = s.encode("ascii")
    return [x >> 5 for x in b] + [0] + [x & 31 for x in b]


def bech32_verify_checksum(hrp, data):
    return bech32_polymod(bech32_hrp_expand(hrp) + data) == 1


def bech32_create_checksum(hrp, data):
    values = bech32_hrp_expand(hrp) + data
    polymod = bech32_polymod(values + [0, 0, 0, 0, 0, 0]) ^ 1
    return [(polymod >> 5 * (5 - i)) & 31 for i in range(6)]


def group_32(s):
    """Convert from 8-bit bytes to 5-bit array of integers"""
    result = []
    unused_bits = 0
    current = 0
    for c in s:
        unused_bits += 8
        current = (current << 8) + c
        while unused_bits > 5:
            unused_bits -= 5
            result.append(current >> unused_bits)
            mask = (1 << unused_bits) - 1
            current &= mask
    result.append(current << (5 - unused_bits))
    return result


def convertbits(data, frombits, tobits, pad=True):
    """General power-of-2 base conversion."""
    acc = 0
    bits = 0
    ret = []
    maxv = (1 << tobits) - 1
    max_acc = (1 << (frombits + tobits - 1)) - 1
    for value in data:
        if value < 0 or (value >> frombits):
            return None
        acc = ((acc << frombits) | value) & max_acc
        bits += frombits
        while bits >= tobits:
            bits -= tobits
            ret.append((acc >> bits) & maxv)
    if pad:
        if bits:
            ret.append((acc << (tobits - bits)) & maxv)
    elif bits >= frombits or ((acc << (tobits - bits)) & maxv):
        return None
    return ret


def bc32encode(data: bytes) -> str:
    """
    bc32 encoding
    see https://github.com/BlockchainCommons/Research/blob/master/papers/bcr-2020-004-bc32.md
    """
    dd = convertbits(data, 8, 5)
    polymod = bech32_polymod([0] + dd + [0, 0, 0, 0, 0, 0]) ^ 0x3FFFFFFF
    chk = [(polymod >> 5 * (5 - i)) & 31 for i in range(6)]
    return "".join([BECH32_ALPHABET[d] for d in dd + chk])


def bc32decode(bc32: str) -> bytes:
    """
    bc32 decoding
    see https://github.com/BlockchainCommons/Research/blob/master/papers/bcr-2020-004-bc32.md
    """
    if bc32.lower() != bc32 and bc32.upper() != bc32:
        return None
    bc32 = bc32.lower()
    if not all([x in BECH32_ALPHABET for x in bc32]):
        return None
    res = [BECH32_ALPHABET.find(c) for c in bc32.lower()]
    if bech32_polymod([0] + res) != 0x3FFFFFFF:
        return None
    return bytes(convertbits(res[:-6], 5, 8, False))


def cbor_encode(data):
    length = len(data)
    if length <= 23:
        prefix = bytes([0x40 + length])
    elif length <= 255:
        prefix = bytes([0x58, length])
    elif length <= 65535:
        prefix = b"\x59" + length.to_bytes(2, "big")
    else:
        prefix = b"\x60" + length.to_bytes(4, "big")
    return prefix + data


def cbor_decode(data):
    s = BytesIO(data)
    b = s.read(1)[0]
    if b >= 0x40 and b < 0x58:
        length = b - 0x40
        return s.read(length)
    if b == 0x58:
        length = s.read(1)[0]
        return s.read(length)
    if b == 0x59:
        length = int.from_bytes(s.read(2), "big")
        return s.read(length)
    if b == 0x60:
        length = int.from_bytes(s.read(4), "big")
        return s.read(length)
    return None


def encode_bech32(nums):
    """Convert from 5-bit array of integers to bech32 format"""
    result = ""
    for n in nums:
        result += BECH32_ALPHABET[n]
    return result


def encode_bech32_checksum(s, network="mainnet"):
    """Convert a segwit ScriptPubKey to a bech32 address"""
    if network == "mainnet":
        prefix = "bc"
    else:
        prefix = "tb"
    version = s[0]
    if version > 0:
        version -= 0x50
    length = s[1]
    data = [version] + group_32(s[2 : 2 + length])
    checksum = bech32_create_checksum(prefix, data)
    bech32 = encode_bech32(data + checksum)
    return prefix + "1" + bech32


def decode_bech32(s):
    """Returns network, segwit version and the hash from the bech32 address"""
    hrp, raw_data = s.split("1")
    if hrp == "bc":
        network = "mainnet"
    elif hrp == "tb":
        network = "testnet"
    else:
        raise ValueError("unknown human readable part: {}".format(hrp))
    data = [BECH32_ALPHABET.index(c) for c in raw_data]
    if not bech32_verify_checksum(hrp, data):
        raise ValueError("bad address: {}".format(s))
    version = data[0]
    number = 0
    for digit in data[1:-6]:
        number = (number << 5) + digit
    num_bytes = (len(data) - 7) * 5 // 8
    bits_to_ignore = (len(data) - 7) * 5 % 8
    number >>= bits_to_ignore
    hash = int_to_big_endian(number, num_bytes)
    if num_bytes < 2 or num_bytes > 40:
        raise ValueError("bytes out of range: {}".format(num_bytes))
    return [network, version, hash]



import hashlib
import hmac

from buidl.helper import (
    big_endian_to_int,
    encode_base58_checksum,
    hash160,
    hash256,
    int_to_big_endian,
    raw_decode_base58,
)
from buidl._libsec import ffi, lib


GLOBAL_CTX = ffi.gc(
    lib.secp256k1_context_create(
        lib.SECP256K1_CONTEXT_SIGN | lib.SECP256K1_CONTEXT_VERIFY
    ),
    lib.secp256k1_context_destroy,
)
P = 2 ** 256 - 2 ** 32 - 977
N = 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFEBAAEDCE6AF48A03BBFD25E8CD0364141


class S256Point:
    def __init__(self, csec=None, usec=None):
        if usec:
            self.usec = usec
            self.csec = None
            sec_cache = usec
        elif csec:
            self.csec = csec
            self.usec = None
            sec_cache = csec
        else:
            raise RuntimeError("need a serialization")
        self.c = ffi.new("secp256k1_pubkey *")
        if not lib.secp256k1_ec_pubkey_parse(
            GLOBAL_CTX, self.c, sec_cache, len(sec_cache)
        ):
            raise ValueError("libsecp256k1 produced error")

    def __eq__(self, other):
        return self.sec() == other.sec()

    def __repr__(self):
        return "S256Point({})".format(self.sec(compressed=False).hex())

    def __rmul__(self, coefficient):
        coef = coefficient % N
        new_key = ffi.new("secp256k1_pubkey *")
        s = self.sec(compressed=False)
        lib.secp256k1_ec_pubkey_parse(GLOBAL_CTX, new_key, s, len(s))
        lib.secp256k1_ec_pubkey_tweak_mul(GLOBAL_CTX, new_key, coef.to_bytes(32, "big"))
        serialized = ffi.new("unsigned char [65]")
        output_len = ffi.new("size_t *", 65)
        lib.secp256k1_ec_pubkey_serialize(
            GLOBAL_CTX, serialized, output_len, new_key, lib.SECP256K1_EC_UNCOMPRESSED
        )
        return self.__class__(usec=bytes(serialized))

    def __add__(self, scalar):
        """Multiplies scalar by generator, adds result to current point"""
        coef = scalar % N
        new_key = ffi.new("secp256k1_pubkey *")
        s = self.sec(compressed=False)
        lib.secp256k1_ec_pubkey_parse(GLOBAL_CTX, new_key, s, len(s))
        lib.secp256k1_ec_pubkey_tweak_add(GLOBAL_CTX, new_key, coef.to_bytes(32, "big"))
        serialized = ffi.new("unsigned char [65]")
        output_len = ffi.new("size_t *", 65)
        lib.secp256k1_ec_pubkey_serialize(
            GLOBAL_CTX, serialized, output_len, new_key, lib.SECP256K1_EC_UNCOMPRESSED
        )
        return self.__class__(usec=bytes(serialized))

    def verify(self, z, sig):
        msg = z.to_bytes(32, "big")
        sig_data = sig.cdata()
        return lib.secp256k1_ecdsa_verify(GLOBAL_CTX, sig_data, msg, self.c)

    def sec(self, compressed=True):
        """returns the binary version of the SEC format"""
        if compressed:
            if not self.csec:
                serialized = ffi.new("unsigned char [33]")
                output_len = ffi.new("size_t *", 33)

                lib.secp256k1_ec_pubkey_serialize(
                    GLOBAL_CTX,
                    serialized,
                    output_len,
                    self.c,
                    lib.SECP256K1_EC_COMPRESSED,
                )
                self.csec = bytes(ffi.buffer(serialized, 33))
            return self.csec
        else:
            if not self.usec:
                serialized = ffi.new("unsigned char [65]")
                output_len = ffi.new("size_t *", 65)

                lib.secp256k1_ec_pubkey_serialize(
                    GLOBAL_CTX,
                    serialized,
                    output_len,
                    self.c,
                    lib.SECP256K1_EC_UNCOMPRESSED,
                )
                self.usec = bytes(ffi.buffer(serialized, 65))
            return self.usec

    def hash160(self, compressed=True):
        # get the sec
        sec = self.sec(compressed)
        # hash160 the sec
        return hash160(sec)

    def p2pkh_script(self, compressed=True):
        """Returns the p2pkh Script object"""
        h160 = self.hash160(compressed)
        # avoid circular dependency
        from buidl.script import P2PKHScriptPubKey

        return P2PKHScriptPubKey(h160)

    def p2wpkh_script(self):
        """Returns the p2wpkh Script object"""
        h160 = self.hash160(True)
        # avoid circular dependency
        from buidl.script import P2WPKHScriptPubKey

        return P2WPKHScriptPubKey(h160)

    def p2sh_p2wpkh_redeem_script(self):
        """Returns the RedeemScript for a p2sh-p2wpkh redemption"""
        return self.p2wpkh_script().redeem_script()

    def address(self, compressed=True, network="mainnet"):
        """Returns the p2pkh address string"""
        return self.p2pkh_script(compressed).address(network)

    def bech32_address(self, network="mainnet"):
        """Returns the p2wpkh bech32 address string"""
        return self.p2wpkh_script().address(network)

    def p2sh_p2wpkh_address(self, network="mainnet"):
        """Returns the p2sh-p2wpkh base58 address string"""
        return self.p2wpkh_script().p2sh_address(network)

    def verify_message(self, message, sig):
        """Verify a message in the form of bytes. Assumes that the z
        is calculated using hash256 interpreted as a big-endian integer"""
        # calculate the hash256 of the message
        h256 = hash256(message)
        # z is the big-endian interpretation. use big_endian_to_int
        z = big_endian_to_int(h256)
        # verify the message using the self.verify method
        return self.verify(z, sig)

    @classmethod
    def parse(self, sec_bin):
        """returns a Point object from a SEC binary (not hex)"""
        if sec_bin[0] == 4:
            return S256Point(usec=sec_bin)
        else:
            return S256Point(csec=sec_bin)


G = S256Point(
    usec=bytes.fromhex(
        "0479be667ef9dcbbac55a06295ce870b07029bfcdb2dce28d959f2815b16f81798483ada7726a3c4655da4fbfc0e1108a8fd17b448a68554199c47d08ffb10d4b8"
    )
)


class Signature:
    def __init__(self, der=None, c=None):
        if der:
            self.der_cache = der
            self.c = ffi.new("secp256k1_ecdsa_signature *")
            if not lib.secp256k1_ecdsa_signature_parse_der(
                GLOBAL_CTX, self.c, der, len(der)
            ):
                raise RuntimeError("badly formatted signature {}".format(der.hex()))
        elif c:
            self.c = c
            self.der_cache = None
        else:
            raise RuntimeError("need der or c object")

    def __eq__(self, other):
        return self.der() == other.der()

    def __repr__(self):
        return "Signature{}".format(self.der().hex())

    def der(self):
        if not self.der_cache:
            der = ffi.new("unsigned char[72]")
            der_length = ffi.new("size_t *", 72)
            lib.secp256k1_ecdsa_signature_serialize_der(
                GLOBAL_CTX, der, der_length, self.c
            )
            self.der_cache = bytes(ffi.buffer(der, der_length[0]))
        return self.der_cache

    def cdata(self):
        return self.c

    @classmethod
    def parse(cls, der):
        return cls(der=der)


class PrivateKey:
    def __init__(self, secret, network="mainnet", compressed=True):
        self.secret = secret
        self.point = secret * G
        self.network = network
        self.compressed = compressed

    def hex(self):
        return "{:x}".format(self.secret).zfill(64)

    def sign(self, z):
        secret = self.secret.to_bytes(32, "big")
        msg = z.to_bytes(32, "big")
        csig = ffi.new("secp256k1_ecdsa_signature *")
        if not lib.secp256k1_ecdsa_sign(
            GLOBAL_CTX, csig, msg, secret, ffi.NULL, ffi.NULL
        ):
            raise RuntimeError("something went wrong with c signing")
        sig = Signature(c=csig)
        if not self.point.verify(z, sig):
            raise RuntimeError("something went wrong with signing")
        return sig

    def deterministic_k(self, z):
        k = b"\x00" * 32
        v = b"\x01" * 32
        if z > N:
            z -= N
        z_bytes = int_to_big_endian(z, 32)
        secret_bytes = int_to_big_endian(self.secret, 32)
        s256 = hashlib.sha256
        k = hmac.new(k, v + b"\x00" + secret_bytes + z_bytes, s256).digest()
        v = hmac.new(k, v, s256).digest()
        k = hmac.new(k, v + b"\x01" + secret_bytes + z_bytes, s256).digest()
        v = hmac.new(k, v, s256).digest()
        while True:
            v = hmac.new(k, v, s256).digest()
            candidate = big_endian_to_int(v)
            if candidate >= 1 and candidate < N:
                return candidate
            k = hmac.new(k, v + b"\x00", s256).digest()
            v = hmac.new(k, v, s256).digest()

    def sign_message(self, message):
        """Sign a message in the form of bytes instead of the z. The z should
        be assumed to be the hash256 of the message interpreted as a big-endian
        integer."""
        # compute the hash256 of the message
        h256 = hash256(message)
        # z is the big-endian interpretation. use big_endian_to_int
        z = big_endian_to_int(h256)
        # sign the message using the self.sign method
        return self.sign(z)

    @classmethod
    def parse(cls, wif):
        """Converts WIF to a PrivateKey object"""
        raw = raw_decode_base58(wif)
        if len(raw) == 34:
            compressed = True
            if raw[-1] != 1:
                raise ValueError("Invalid WIF")
            raw = raw[:-1]
        else:
            compressed = False
        secret = big_endian_to_int(raw[1:])
        if raw[0] == 0xEF:
            network = "testnet"
        elif raw[0] == 0x80:
            network = "mainnet"
        else:
            raise ValueError("Invalid WIF")
        return cls(secret, network=network, compressed=compressed)

    def wif(self, compressed=True):
        # convert the secret from integer to a 32-bytes in big endian using num.to_bytes(32, 'big')
        secret_bytes = self.secret.to_bytes(32, "big")
        # prepend b'\xef' on testnet, b'\x80' on mainnet
        if self.network == "mainnet":
            prefix = b"\x80"
        else:
            prefix = b"\xef"
        # append b'\x01' if compressed
        if compressed:
            suffix = b"\x01"
        else:
            suffix = b""
        # encode_base58_checksum the whole thing
        return encode_base58_checksum(prefix + secret_bytes + suffix)



from setuptools import setup, find_packages


with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="buidl",
    version="0.2.27",
    author="Example Author",
    author_email="author@example.com",
    description="An easy-to-use and fully featured bitcoin library written in pure python (no dependencies).",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/buidl-bitcoin/buidl-python",
    packages=find_packages(),
    include_package_data=True,  # https://stackoverflow.com/a/56689053
    scripts=["multiwallet.py", "singlesweep.py"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)



from unittest import TestCase
from copy import deepcopy

from buidl.hd import HDPrivateKey, HDPublicKey
from buidl.psbt_helper import create_p2sh_multisig_psbt
from buidl.script import RedeemScript


class P2SHTest(TestCase):
    def test_receive_1of2(self):
        # This test is not strictly neccesary, it just proves/shows how I generated the testnet address that received these coins
        # In the next test, I'll show to spend them

        # For full details, see test_create_p2sh_multisig in test_script.py

        # Insecure testing BIP39 mnemonics
        root_path = "m/45'/0/0/0"
        hdprivs = [
            HDPrivateKey.from_mnemonic(seed_word * 12, network="testnet")
            for seed_word in ("action ", "agent ")
        ]

        # Validate the 0th receive address for m/45'/0
        expected_spending_addr = "2ND4qfpdHyeXJboAUkKZqJsyiKyXvHRKhbi"
        pubkey_hexes = [
            hdpriv.traverse(root_path).pub.sec().hex() for hdpriv in hdprivs
        ]

        redeem_script_to_use = RedeemScript.create_p2sh_multisig(
            quorum_m=1,
            pubkey_hexes=pubkey_hexes,
            sort_keys=True,
        )
        self.assertEqual(
            expected_spending_addr, redeem_script_to_use.address(network="testnet")
        )

        # Faucet TX sent to this address:
        # https://blockstream.info/testnet/tx/4412d2a7664d01bb784a0a359e9aacf160ee436067c6a42dca355da4817ca7da

    def test_sweep_1of2_p2sh(self):

        # This test will produce a validly signed TX for the 1-of-2 p2sh using either key, which will result in a different TX ID
        # d8172be9981a4f57e6e4ebe0f4785f5f2035aee40ffbb2d6f1200810a879d490 is the one that was broadcast to the testnet blockchain:
        # https://blockstream.info/testnet/tx/d8172be9981a4f57e6e4ebe0f4785f5f2035aee40ffbb2d6f1200810a879d490

        kwargs = {
            "public_key_records": [
                # action x12
                [
                    "e0c595c5",
                    "tpubDBnspiLZfrq1V7j1iuMxGiPsuHyy6e4QBnADwRrbH89AcnsUEMfWiAYXmSbMuNFsrMdnbQRDGGSM1AFGL6zUWNVSmwRavoJzdQBbZKLgLgd",
                    "m/45'/0",
                ],
                # agent x12
                [
                    "838f3ff9",
                    "tpubDAKJicb9Tkw34PFLEBUcbnH99twN3augmg7oYHHx9Aa9iodXmA4wtGEJr8h2XjJYqn2j1v5qHLjpWEe8aPihmC6jmsgomsuc9Zeh4ZushNk",
                    "m/45'/0",
                ],
            ],
            "input_dicts": [
                {
                    "quorum_m": 1,
                    "path_dict": {
                        # xfp: root_path
                        "e0c595c5": "m/45'/0/0/0",
                        "838f3ff9": "m/45'/0/0/0",
                    },
                    "prev_tx_dict": {
                        "hex": "02000000000101380bff9db676d159ad34849079c77e0d5c1df9c841b6a6640cba9bfc15077eea0100000000feffffff02008312000000000017a914d96bb9c5888f473dbd077d77009fb49ba2fda24287611c92a00100000017a9148722f07fbcf0fc506ea4ba9daa811d11396bbcfd870247304402202fe3c2f18e1486407bf0baabd2b3376102f0844a754d8e2fb8de71b39b3f76c702200c1fe8f7f9ef5165929ed51bf754edd7dd3e591921979cf5b891c841a1fd19d80121037c8fe1fa1ae4dfff522c532917c73c4884469e3b6a284e9a039ec612dca78eefd29c1e00",
                        "hash_hex": "4412d2a7664d01bb784a0a359e9aacf160ee436067c6a42dca355da4817ca7da",
                        "output_idx": 0,
                        "output_sats": 1213184,
                    },
                },
            ],
            "output_dicts": [
                {
                    "sats": 999500,
                    "address": "mkHS9ne12qx9pS9VojpwU5xtRd4T7X7ZUt",
                },
            ],
            "fee_sats": 213684,
        }

        expected_unsigned_psbt_b64 = "cHNidP8BAFUBAAAAAdqnfIGkXTXKLaTGZ2BD7mDxrJqeNQpKeLsBTWan0hJEAAAAAAD/////AUxADwAAAAAAGXapFDRKD0jKFQ7CuQOBdmC5tosTpnAmiKwAAAAATwEENYfPAheO+c4AAAAALjD5unTcevzcKaWlkg/+xoU+FD5bJ+iltDa2WE1bTngCfSOd3kkScA1e4OGM3MJ/Oqg+nxEHlwuV7YBCuoT745YMg48/+S0AAIAAAAAATwEENYfPAuBIVlUAAAAAHnpWPWgBnfWu2tzv9Ujws27ps0hOBHMQbZlpQ6c5m4MDN4/ffnc3ejfVlqkEgG7tlPX4aTS92pfU5LAvGDQQKHoM4MWVxS0AAIAAAAAAAAEA4AIAAAAAAQE4C/+dtnbRWa00hJB5x34NXB35yEG2pmQMupv8FQd+6gEAAAAA/v///wIAgxIAAAAAABepFNlrucWIj0c9vQd9dwCftJui/aJCh2EckqABAAAAF6kUhyLwf7zw/FBupLqdqoEdETlrvP2HAkcwRAIgL+PC8Y4UhkB78Lqr0rM3YQLwhEp1TY4vuN5xs5s/dscCIAwf6Pf571Flkp7VG/dU7dfdPlkZIZec9biRyEGh/RnYASEDfI/h+hrk3/9SLFMpF8c8SIRGnjtqKE6aA57GEtynju/SnB4AAQRHUSECE59xZ/F9qU0k32/ihozbXtq9DP+Dq0h8lC6i8HVw2fAhAjy3HGmZkHaMAY3xejZsTrdL3hafKeiEJhwQyXPsJq06Uq4iBgITn3Fn8X2pTSTfb+KGjNte2r0M/4OrSHyULqLwdXDZ8BSDjz/5LQAAgAAAAAAAAAAAAAAAACIGAjy3HGmZkHaMAY3xejZsTrdL3hafKeiEJhwQyXPsJq06FODFlcUtAACAAAAAAAAAAAAAAAAAAAA="
        tests = (
            # (seed_word repeated x12, signed_tx_hash_hex),
            (
                # this is the one we broadcasted
                "action ",
                "d8172be9981a4f57e6e4ebe0f4785f5f2035aee40ffbb2d6f1200810a879d490",
            ),
            (
                "agent ",
                "c38dcee54f40193d668c8911eeba7ec20f570fdbdb31fbe4351f66d7005c7bfb",
            ),
        )

        # Now we prove we can sign this with either key
        for seed_word, signed_tx_hash_hex in tests:
            psbt_obj = create_p2sh_multisig_psbt(**kwargs)
            self.assertEqual(len(psbt_obj.hd_pubs), 2)
            self.assertEqual(psbt_obj.serialize_base64(), expected_unsigned_psbt_b64)

            hdpriv = HDPrivateKey.from_mnemonic(seed_word * 12, network="testnet")

            root_path_to_use = None
            for cnt, psbt_in in enumerate(psbt_obj.psbt_ins):

                self.assertEqual(psbt_in.redeem_script.get_quorum(), (1, 2))

                # For this TX there is only one psbt_in (1 input)
                for child_pubkey in psbt_in.redeem_script.signing_pubkeys():
                    named_pubkey = psbt_in.named_pubs[child_pubkey]
                    if (
                        named_pubkey.root_fingerprint.hex()
                        == hdpriv.fingerprint().hex()
                    ):
                        root_path_to_use = named_pubkey.root_path

                # In this example, the path is the same regardless of which key we sign with:
                self.assertEqual(root_path_to_use, "m/45'/0/0/0")

            private_keys = [hdpriv.traverse(root_path_to_use).private_key]

            self.assertTrue(psbt_obj.sign_with_private_keys(private_keys=private_keys))

            psbt_obj.finalize()

            self.assertEqual(psbt_obj.final_tx().hash().hex(), signed_tx_hash_hex)

    def test_spend_1of2_with_change(self):

        kwargs = {
            # this part is unchanged from the previous
            "public_key_records": [
                # action x12
                [
                    "e0c595c5",
                    "tpubDBnspiLZfrq1V7j1iuMxGiPsuHyy6e4QBnADwRrbH89AcnsUEMfWiAYXmSbMuNFsrMdnbQRDGGSM1AFGL6zUWNVSmwRavoJzdQBbZKLgLgd",
                    "m/45'/0",
                ],
                # agent x12
                [
                    "838f3ff9",
                    "tpubDAKJicb9Tkw34PFLEBUcbnH99twN3augmg7oYHHx9Aa9iodXmA4wtGEJr8h2XjJYqn2j1v5qHLjpWEe8aPihmC6jmsgomsuc9Zeh4ZushNk",
                    "m/45'/0",
                ],
            ],
            # this part is changed:
            "input_dicts": [
                {
                    "quorum_m": 1,
                    "path_dict": {
                        # xfp: root_path
                        "e0c595c5": "m/45'/0/0/1",
                        "838f3ff9": "m/45'/0/0/1",
                    },
                    "prev_tx_dict": {
                        "hex": "020000000001012c40a6810f7a670913d171e1f5b203ca01ed45ed3bf68b649850491eecb560080100000000feffffff02a7e50941010000001600147b3af2253632c3000f9cdd531747107fe249c7d1102700000000000017a91459fb638aaa55a7119a09faf5e8b2ce8a879cce338702473044022004666d885310990e1b0a61e93b1490acb172d43200d6fcfa22e89905b7f3094d02204a705e4a4fc8cab97f7d146f481e70718ee4567486796536d14c7808be3fd866012102cc3b01d2192b5275d3fda7f82eaf593dfb8ca9333f7296f93da401f8d1821335619f1e00",
                        "hash_hex": "3bbc91c1de188528d254ed145b1aabbb68b9050f7c660b345fc9feba73ff94a2",
                        "output_idx": 1,
                        "output_sats": 10000,
                    },
                },
            ],
            "output_dicts": [
                {
                    # this should be change:
                    "sats": 1000,
                    "address": "2MzQhXqN93igSKGW9CMvkpZ9TYowWgiNEF8",
                    "quorum_m": 1,
                    "path_dict": {
                        # xfp: root_path (m/.../1/*/{idx} is change addr branch)
                        "e0c595c5": "m/45'/0/1/0",
                        "838f3ff9": "m/45'/0/1/0",
                    },
                },
                {
                    # testnet faucet:
                    "sats": 5000,
                    "address": "tb1ql7w62elx9ucw4pj5lgw4l028hmuw80sndtntxt",
                },
            ],
            "fee_sats": 4000,
        }

        expected_unsigned_psbt_b64 = "cHNidP8BAHIBAAAAAaKU/3O6/slfNAtmfA8FuWi7qxpbFO1U0iiFGN7Bkbw7AQAAAAD/////AugDAAAAAAAAF6kUTpOd3+VnlxaS0ZVZ8fktT65aoFaHiBMAAAAAAAAWABT/naVn5i8w6oZU+h1fvUe++OO+EwAAAABPAQQ1h88CF475zgAAAAAuMPm6dNx6/NwppaWSD/7GhT4UPlsn6KW0NrZYTVtOeAJ9I53eSRJwDV7g4Yzcwn86qD6fEQeXC5XtgEK6hPvjlgyDjz/5LQAAgAAAAABPAQQ1h88C4EhWVQAAAAAeelY9aAGd9a7a3O/1SPCzbumzSE4EcxBtmWlDpzmbgwM3j99+dzd6N9WWqQSAbu2U9fhpNL3al9TksC8YNBAoegzgxZXFLQAAgAAAAAAAAQDfAgAAAAABASxApoEPemcJE9Fx4fWyA8oB7UXtO/aLZJhQSR7stWAIAQAAAAD+////AqflCUEBAAAAFgAUezryJTYywwAPnN1TF0cQf+JJx9EQJwAAAAAAABepFFn7Y4qqVacRmgn69eiyzoqHnM4zhwJHMEQCIARmbYhTEJkOGwph6TsUkKyxctQyANb8+iLomQW38wlNAiBKcF5KT8jKuX99FG9IHnBxjuRWdIZ5ZTbRTHgIvj/YZgEhAsw7AdIZK1J10/2n+C6vWT37jKkzP3KW+T2kAfjRghM1YZ8eAAEER1EhAiblPHq2FdpGg2TKn1mHn1G9lCeMJqOcrFmiNafPgUWTIQLSIeF7uAzxMLsJMUNK+sXlXzd+2bKwA1GaUpzgKGJER1KuIgYCJuU8erYV2kaDZMqfWYefUb2UJ4wmo5ysWaI1p8+BRZMUg48/+S0AAIAAAAAAAAAAAAEAAAAiBgLSIeF7uAzxMLsJMUNK+sXlXzd+2bKwA1GaUpzgKGJERxTgxZXFLQAAgAAAAAAAAAAAAQAAAAABAEdRIQJ3tPielqw9CAqDy+Tn23vo4cUxiYYGk/KWtw5ryOPJ3yEC1WIbkoA7VEDNWw0VmAntyxp6EY5IfZ7sbF1PNJklWFFSriICAne0+J6WrD0ICoPL5Ofbe+jhxTGJhgaT8pa3DmvI48nfFODFlcUtAACAAAAAAAEAAAAAAAAAIgIC1WIbkoA7VEDNWw0VmAntyxp6EY5IfZ7sbF1PNJklWFEUg48/+S0AAIAAAAAAAQAAAAAAAAAAAA=="

        # With p2sh, txid changes depending on which key signs
        tests = (
            # (seed_word repeated x12, signed_tx_hash_hex),
            (
                # this one we did not broadcast
                "action ",
                "0eefb9c614abff0fe859c2dd524a5cfc9582389c9ec938c02f2afb36c64a8e69",
            ),
            (
                # this is the one we did broadcast
                # https://blockstream.info/testnet/tx/5b7f81bbef354a48097d429cc6e5b7aad1a1b6940faa4aba284a8913fff643dc
                "agent ",
                "5b7f81bbef354a48097d429cc6e5b7aad1a1b6940faa4aba284a8913fff643dc",
            ),
        )

        # Now we prove we can sign this with either key
        for seed_word, signed_tx_hash_hex in tests:
            psbt_obj = create_p2sh_multisig_psbt(**kwargs)
            self.assertEqual(len(psbt_obj.hd_pubs), 2)
            self.assertEqual(psbt_obj.serialize_base64(), expected_unsigned_psbt_b64)

            hdpriv = HDPrivateKey.from_mnemonic(seed_word * 12, network="testnet")

            root_path_to_use = None
            for cnt, psbt_in in enumerate(psbt_obj.psbt_ins):

                self.assertEqual(psbt_in.redeem_script.get_quorum(), (1, 2))

                # For this TX there is only one psbt_in (1 input)
                for child_pubkey in psbt_in.redeem_script.signing_pubkeys():
                    named_pubkey = psbt_in.named_pubs[child_pubkey]
                    if (
                        named_pubkey.root_fingerprint.hex()
                        == hdpriv.fingerprint().hex()
                    ):
                        root_path_to_use = named_pubkey.root_path

                # In this example, the path is the same regardless of which key we sign with:
                self.assertEqual(root_path_to_use, "m/45'/0/0/1")

            private_keys = [hdpriv.traverse(root_path_to_use).private_key]

            self.assertTrue(psbt_obj.sign_with_private_keys(private_keys=private_keys))

            psbt_obj.finalize()

            self.assertEqual(psbt_obj.final_tx().hash().hex(), signed_tx_hash_hex)

        # Replace xfps
        psbt_obj = create_p2sh_multisig_psbt(**kwargs)
        with self.assertRaises(ValueError) as cm:
            # deadbeef  not in psbt
            psbt_obj.replace_root_xfps({"deadbeef": "00000000"})
        self.assertEqual(str(cm.exception), "xfp_hex deadbeef not found in psbt")

        psbt_obj.replace_root_xfps({"e0c595c5": "00000000"})
        self.assertNotEqual(psbt_obj.serialize_base64(), expected_unsigned_psbt_b64)

        # Confirm that swapping out the change address throws an error
        # nonsense address corresponding to secret exponent = 1
        modified_kwargs = deepcopy(kwargs)
        fake_addr = "tb1qw508d6qejxtdg4y5r3zarvary0c5xw7kxpjzsx"
        modified_kwargs["output_dicts"][0]["address"] = fake_addr
        with self.assertRaises(ValueError) as cm:
            create_p2sh_multisig_psbt(**modified_kwargs)
        self.assertEqual(
            "Invalid redeem script for output #0. Expecting 2MzQhXqN93igSKGW9CMvkpZ9TYowWgiNEF8 but got tb1qw508d6qejxtdg4y5r3zarvary0c5xw7kxpjzsx",
            str(cm.exception),
        )

        # Confirm that changing the change paths (so that they no longer match the change address) throws an error
        # Change the path used to validate the change address
        # We don't bother altering 838f3ff9's path as well because changing anything will already throw this error
        modified_kwargs = deepcopy(kwargs)
        modified_kwargs["output_dicts"][0]["path_dict"]["e0c595c5"] = "m/999"
        with self.assertRaises(ValueError) as cm:
            create_p2sh_multisig_psbt(**modified_kwargs)
        self.assertEqual(
            "xfp_hex e0c595c5 with m/999 for in/output #0 not supplied in xpub_dict",
            str(cm.exception),
        )

    def test_full_wallet_functionality(self):
        """
        Create a wallet, fund it, and spend it (validating on all devices first).
        Do this both with and without change.
        """

        child_path = "m/0/0"
        base_path = "m/45'/0"
        xpubs = [
            # Verified against electrum 2021-10
            "tpubDAp5uHrhyCRnoQdMmF9xFU78bev55TKux39Viwn8VCnWywjTCBB3TFSA1i5i4scj8ZkD8RR37EZMPyqBwk4wD4VNpxyowgk2rQ7i8AdVauN",
            "tpubDBxHJMmnouRSes7kweA5wUp2Jm5F1xEPCG3fjYRnBSiQM7hjsp9gp8Wsag16A8yrpMgJkPX8iddXmwuxKaUnAWfdcGcuXbiATYeGXCKkEXK",
            "tpubDANTttKyukGYwXvhTmdCa6p1nHmsbFef2yRXb9b2JTvLGoekYk6ZG9oHXGHYJjTn5g3DijvQrZ4rH6EosYqpVngKiUx2jvWdPmq6ZSDr3ZE",
        ]

        xfps = ["9a6a2580", "8c7e2500", "0586b1ce"]

        pubkey_records = []
        for cnt, seed_word in enumerate(("bacon", "flag", "gas")):
            seed_phrase = f"{seed_word} " * 24
            hd_obj = HDPrivateKey.from_mnemonic(seed_phrase, network="testnet")

            # takes the form xfp, child_xpub, base_path
            pubkey_record = [
                hd_obj.fingerprint().hex(),
                hd_obj.traverse(base_path).xpub(),
                base_path,
            ]
            pubkey_records.append(pubkey_record)

            self.assertEqual(pubkey_record, [xfps[cnt], xpubs[cnt], base_path])

        pubkey_hexes = [
            HDPublicKey.parse(tpub).traverse(child_path).sec().hex() for tpub in xpubs
        ]

        rs_obj = RedeemScript.create_p2sh_multisig(
            quorum_m=2, pubkey_hexes=pubkey_hexes, sort_keys=True
        )

        deposit_address = rs_obj.address(network="testnet")
        # We funded this testnet address in 2021-10:
        self.assertEqual(deposit_address, "2MzLzwPgo6ZTyPzkguNWKfFHgCXdZSJu9tL")

        input_dicts = [
            {
                "quorum_m": 2,
                "path_dict": {
                    # xfp: root_path
                    "9a6a2580": "m/45'/0/0/0",
                    "8c7e2500": "m/45'/0/0/0",
                    "0586b1ce": "m/45'/0/0/0",
                },
                "prev_tx_dict": {
                    # This was the funding transaction send to 2MzLzwPgo6ZTyPzkguNWKfFHgCXdZSJu9tL
                    "hex": "02000000000101045448b2faafc53a7586bd6a746a598d9a1cfc3dd548d8f8afd12c11c53c16740000000000feffffff02809a4b000000000017a914372becc78d20f8acb94059fa8d1e3219b67c1d9387a08601000000000017a9144de07b9788bc64143292f9610443a7f81cc5fc308702473044022059f2036e5b6da03e592863664082f957472dbb89330895d23ac66f94e3819f63022050d31d401cd86fe797c31ba7fb2a2cf9e12356bba8388897f4f522bc4999224f0121029c71554e111bb41463ad288b4808effef8136755da80d42aa2bcea12ed8a99bbba0e2000",
                    "hash_hex": "838504ad934d97915d9ab58b0ece6900ed123abf3a51936563ba9d67b0e41fa4",
                    "output_idx": 1,
                    "output_sats": 100_000,
                },
            }
        ]

        base_description_want = {
            "locktime": 0,
            "version": 1,
            "network": "testnet",
            "spend_addr": "2MxvTasFXms6w1R6omwqtKKTYDyjrbhCTYd",
            "is_batch_tx": False,
            "inputs_desc": [
                {
                    "quorum": "2-of-3",
                    "bip32_derivs": [
                        {
                            "pubkey": "02861f9fd125b4183cd3d5645ff7c7697573e7cf6da174aa993a25d498191f4ff4",
                            "master_fingerprint": "0586b1ce",
                            "path": "m/45'/0/0/0",
                            "xpub": "tpubDANTttKyukGYwXvhTmdCa6p1nHmsbFef2yRXb9b2JTvLGoekYk6ZG9oHXGHYJjTn5g3DijvQrZ4rH6EosYqpVngKiUx2jvWdPmq6ZSDr3ZE",
                        },
                        {
                            "pubkey": "03c46aea215fd0017576534cfad56aa06d14043947add8ffe4cb7b4ba1d9b023b3",
                            "master_fingerprint": "8c7e2500",
                            "path": "m/45'/0/0/0",
                            "xpub": "tpubDBxHJMmnouRSes7kweA5wUp2Jm5F1xEPCG3fjYRnBSiQM7hjsp9gp8Wsag16A8yrpMgJkPX8iddXmwuxKaUnAWfdcGcuXbiATYeGXCKkEXK",
                        },
                        {
                            "pubkey": "03f25b39eec4d2824883b2e4cb010adb7685f13540805767c763808a599fd8df36",
                            "master_fingerprint": "9a6a2580",
                            "path": "m/45'/0/0/0",
                            "xpub": "tpubDAp5uHrhyCRnoQdMmF9xFU78bev55TKux39Viwn8VCnWywjTCBB3TFSA1i5i4scj8ZkD8RR37EZMPyqBwk4wD4VNpxyowgk2rQ7i8AdVauN",
                        },
                    ],
                    "prev_txhash": "838504ad934d97915d9ab58b0ece6900ed123abf3a51936563ba9d67b0e41fa4",
                    "prev_idx": 1,
                    "n_sequence": 4294967295,
                    "sats": 100000,
                    "addr": "2MzLzwPgo6ZTyPzkguNWKfFHgCXdZSJu9tL",
                    "witness_script": None,
                    "redeem_script": "OP_2 02861f9fd125b4183cd3d5645ff7c7697573e7cf6da174aa993a25d498191f4ff4 03c46aea215fd0017576534cfad56aa06d14043947add8ffe4cb7b4ba1d9b023b3 03f25b39eec4d2824883b2e4cb010adb7685f13540805767c763808a599fd8df36 OP_3 OP_CHECKMULTISIG ",
                }
            ],
            "root_paths": {
                "0586b1ce": {"m/45'/0/0/0"},
                "8c7e2500": {"m/45'/0/0/0"},
                "9a6a2580": {"m/45'/0/0/0"},
            },
        }

        # We broadcast this to the testnet blockchain (confirmed in block 000000000000000ce65761d5949a930daa84381a5fb86c930947ea88742c957d)
        # Takes the form: (tx_output_dicts, psbt_outputs_desc, psbt_desc_want, expected_tx_hash)
        sweep_test_output = (
            # tx_output_dicts
            [
                # Sweep to faucet
                {
                    "sats": 60_000,
                    "address": "2MxvTasFXms6w1R6omwqtKKTYDyjrbhCTYd",
                },
            ],
            # psbt_outputs_desc
            [
                {
                    "sats": 60_000,
                    "addr": "2MxvTasFXms6w1R6omwqtKKTYDyjrbhCTYd",
                    "addr_type": "P2SH",
                    "is_change": False,
                }
            ],
            # psbt_desc_want
            {
                "txid": "1b768c9e7bc19898fbe92d28a5b3ed482ea9719cd34a77b8b9bcb050a99466c7",
                "tx_fee_sats": 40_000,
                "total_input_sats": 100_000,
                "total_output_sats": 60_000,
                "spend_sats": 60_000,
                "change_addr": "",
                "change_sats": 0,
                "tx_summary_text": "PSBT sends 60,000 sats to 2MxvTasFXms6w1R6omwqtKKTYDyjrbhCTYd with a fee of 40,000 sats (40.0% of spend)",
            },
            # expected_tx_hash
            "229fd99a2abab6c2b193bde547920424d2903978c4129472b06c9d2ffaced025",
        )

        # We sign but don't broadcast this
        # Takes the form: (tx_output_dicts, psbt_outputs_desc, psbt_desc_want, expected_tx_hash)
        change_test_outputs = (
            # tx_output_dicts
            [
                # Spend to faucet
                {
                    "sats": 50_000,
                    "address": "2MxvTasFXms6w1R6omwqtKKTYDyjrbhCTYd",
                },
                # Change back to self
                {
                    "sats": 30_000,
                    "address": "2MzJEYwRjTcpdpMa9J2sb9QQAWNB34umhbV",
                    "quorum_m": 2,
                    "path_dict": {
                        # xfp: root_path (m/.../1/*/{idx} is change addr branch)
                        "0586b1ce": "m/45'/0/1/0",
                        "8c7e2500": "m/45'/0/1/0",
                        "9a6a2580": "m/45'/0/1/0",
                    },
                },
            ],
            # psbt_outputs_desc
            [
                {
                    "sats": 50_000,
                    "addr": "2MxvTasFXms6w1R6omwqtKKTYDyjrbhCTYd",
                    "addr_type": "P2SH",
                    "is_change": False,
                },
                {
                    "sats": 30_000,
                    "addr": "2MzJEYwRjTcpdpMa9J2sb9QQAWNB34umhbV",
                    "addr_type": "P2SH",
                    "is_change": True,
                    "witness_script": None,
                    "redeem_script": "OP_2 02a1921eba209b568f6fa904c4b1e1b87790d2b73d48607aad4db91f977879a16b 02bae8e95f63062500fa18a5142e7103d1783cae028c56293735f7bddbf3ba7b15 032953df01093cab0c2d73067c90509039fe3599a2f80e9b90a6c2229eba432208 OP_3 OP_CHECKMULTISIG ",
                },
            ],
            # psbt_desc_want
            {
                "txid": "076baed6061d88f2342b57e8c9b1b96cf155cec0dce93e6f7d80b2c540876166",
                "tx_fee_sats": 20_000,
                "total_input_sats": 100_000,
                "total_output_sats": 80_000,
                "spend_sats": 50_000,
                "change_addr": "2MzJEYwRjTcpdpMa9J2sb9QQAWNB34umhbV",
                "change_sats": 30_000,
                "tx_summary_text": "PSBT sends 50,000 sats to 2MxvTasFXms6w1R6omwqtKKTYDyjrbhCTYd with a fee of 20,000 sats (20.0% of spend)",
            },
            # expected_tx_hash
            "668a7d61aa09c01db90a477e6aabdc433712b161cf419d23e94f0c1fc3a89097",
        )

        test_outputs = (change_test_outputs, sweep_test_output)

        for (
            tx_output_dicts,
            psbt_outputs_desc,
            psbt_desc_want,
            expected_tx_hash,
        ) in test_outputs:

            # Return all the funds to the faucet address
            psbt_obj = create_p2sh_multisig_psbt(
                public_key_records=pubkey_records,
                input_dicts=input_dicts,
                output_dicts=tx_output_dicts,
                fee_sats=psbt_desc_want["tx_fee_sats"],
            )

            self.assertTrue(psbt_obj.validate())

            description_want = {
                **base_description_want,
                **psbt_desc_want,
                **{"outputs_desc": psbt_outputs_desc},
            }

            self.assertEqual(psbt_obj.describe_basic_multisig(), description_want)

            # choosing to sign with these two seeds, could have signed with any 2-of-3
            private_child_keys = []
            for seed_word in ("bacon", "gas"):
                seed_phrase = f"{seed_word} " * 24
                hd_obj = HDPrivateKey.from_mnemonic(seed_phrase, network="testnet")
                private_child_key = hd_obj.traverse(f"{base_path}/0/0").private_key
                private_child_keys.append(private_child_key)

            self.assertTrue(psbt_obj.sign_with_private_keys(private_child_keys))

            psbt_obj.finalize()

            self.assertTrue(psbt_obj.validate())

            final_tx = psbt_obj.final_tx()

            # We would broadcast this to the network:
            # final_tx.serialize().hex()

            self.assertEqual(final_tx.hash().hex(), expected_tx_hash)
